from discminer.core import Data
from discminer.disc2d import Model
import discminer.cart as cart

import numpy as np
from astropy import units as u

import json
import copy

from argparse import ArgumentParser
from utils import add_parser_args

parser = ArgumentParser(prog='make moments', description='Make double gauss or double bell moments and save outputs into .fits files')
parser.add_argument('-ni', '--niter', default=10, type=int,
                    help="Number of iterations to re-do fit on hot pixels. DEFAULS to 10")
parser.add_argument('-ne', '--neighs', default=5, type=int,
                    help="Number of neighbour pixels on each side of hot pixel considered for the iterative fit. DEFAULS to 5")
args = add_parser_args(parser, kernel='doublebell', kind=True, sigma=True)

with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc
Rmax = 1.3*best['intensity']['Rout']*u.au #Max model radius, 30% larger than disc Rout

#*********
#LOAD DATA
#*********
datacube = Data(file_data, dpc) # Read data and convert to Cube object
vchannels = datacube.vchannels

#****************************
#INIT MODEL AND PRESCRIPTIONS
#****************************
model = Model(datacube, Rmax, Rmin=0, prototype=True)

model.velocity_func = model.keplerian_vertical # vrot = sqrt(GM/r**3)*R
model.line_profile = model.line_profile_bell

if 'sum' in meta['kind']:
    model.line_uplow = model.line_uplow_sum
else:
    model.line_uplow = model.line_uplow_mask

if 'I2pwl' in meta['kind']:
    model.intensity_func = cart.intensity_powerlaw_rbreak
elif 'I2pwlnosurf' in meta['kind']:
    model.intensity_func = cart.intensity_powerlaw_rbreak_nosurf    
else:
    model.intensity_func = cart.intensity_powerlaw_rout

if 'surf2pwl' in meta['kind']:
    model.z_upper_func = cart.z_upper_powerlaw
    model.z_lower_func = cart.z_lower_powerlaw
else:
    model.z_upper_func = cart.z_upper_exp_tapered
    model.z_lower_func = cart.z_lower_exp_tapered
  
#**************
#PROTOTYPE PARS
#**************
model.params = copy.copy(best)
model.params['intensity']['I0'] /= meta['downsamp_factor']
model.make_model()

#**********
#LOAD MODEL
#**********
datacube.convert_to_tb(writefits=False)
file_model = 'cube_model_%s_convtb.fits'%tag
modelcube = Data(file_model, dpc) # Read model and convert to Cube object

#**********************
#MAKE MOMENT MAPS
#**********************
#Use model priors + kernel
moments_data = datacube.make_moments(model=model, method=args.kernel, kind=args.kind, sigma_thres=args.sigma, niter=args.niter, neighs=args.neighs,
                                     writecomp=True, parcube=True, tag='_data') 
moments_model = modelcube.make_moments(model=model, method=args.kernel, kind=args.kind, sigma_thres=0, niter=5, #set sigma 0 to prevent masking of model
                                       writecomp=True, parcube=True, tag='_model') 




from discminer.mining_control import _mining_gradient
from discminer.mining_utils import (load_disc_grid,
                                    init_data_and_model,
                                    get_noise_mask,
                                    get_2d_plot_decorators,
                                    mark_planet_location,
                                    show_output,
                                    load_moments)

import discminer.cart as cart
from discminer.plottools import (make_polar_map,
                                 make_substructures,
                                 make_up_ax,
                                 mod_major_ticks,
                                 mod_nticks_cbars,
                                 make_1d_legend,
                                 use_discminer_style)

import json
import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u

use_discminer_style()

if __name__ == '__main__':
    parser = _mining_gradient(None)
    args = parser.parse_args()

args.projection = 'polar'

#**************************
#JSON AND SOME DEFINITIONS
#**************************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

vsys = best['velocity']['vsys']
Rout = best['intensity']['Rout']
incl = best['orientation']['incl']
PA = best['orientation']['PA']
xc = best['orientation']['xc']
yc = best['orientation']['yc']

gaps = custom['gaps']
rings = custom['rings']

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment, unit_simple=True, fmt_vertical=True)

levels_resid = np.linspace(-clim, clim, 48)

clim_grad = 0.04*clim #For gradient: dv/dR [unit/au]
cmap_grad = plt.get_cmap('nipy_spectral')
cmap_grad_r = plt.get_cmap('nipy_spectral_r')

levels_grad = np.linspace(0.05*clim_grad, clim_grad, 48)

file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc
Rmax = 1.1*Rout*u.au #Max model radius, 10% larger than disc Rout

#*******************
#LOAD DATA AND MODEL
#*******************
datacube, model = init_data_and_model()

noise_mean, mask = get_noise_mask(datacube, thres=2)
vchannels = datacube.vchannels
model.make_model()

R, phi, z = load_disc_grid()

#*****************
#LOAD MOMENT MAPS    
#*****************
if args.absolute_Rinner>=0:
    Rin_plt = args.absolute_Rinner
else:
    Rin_plt = args.Rinner*datacube.beam_size

if args.absolute_Router>=0:
    Rout_plt = args.absolute_Router*u.au
else:
    Rout_plt = args.Router*Rout*u.au

moment_data, moment_model, residuals, mtags = load_moments(
    args,
    mask=mask,
    clip_Rmin=0.0*u.au,
    clip_Rmax=Rout*u.au,
    clip_Rgrid=R[args.surface]*u.m
)

#*******************
#SURFACE PROPERTIES
#*******************
if args.surface in ['up', 'upper']:
    if 'surf2pwl' in meta['kind']:
        z_func = cart.z_upper_powerlaw
    elif 'surfirregular' in meta['kind']:
        z_func = cart.z_upper_irregular
    else:
        z_func = cart.z_upper_exp_tapered
    z_pars = best['height_upper']

elif args.surface in ['low', 'lower']:
    if 'surf2pwl' in meta['kind']:
        z_func = cart.z_lower_powerlaw
    elif 'surfirregular' in meta['kind']:
        z_func = cart.z_lower_irregular            
    else:
        z_func = cart.z_lower_exp_tapered
    z_pars = best['height_lower']

#**************
#FIGURE LABELS
#**************
if args.moment in ['velocity', 'v0phi', 'v0r', 'v0z', 'vr_leftover', 'linewidth']:
    unit = 'm/s'
    ufac = u.Unit('km/s').to(unit)
    dfmt = '%3d'
elif args.moment=='peakintensity':
    unit = 'K'
    ufac = 1
    dfmt = '%4.1f'
else:
    unit = ''
    ufac = 1
    dfmt = '%4.1f'
    
clabels = {
    'linewidth': r'$\Delta$ Line width [m s$^{-1}$]',
    'lineslope': r'$\Delta$ Line slope',
    'velocity': r'$\Delta$ Centroid [m s$^{-1}$]',
    'v0phi': r'$\Delta$ Centroid [m s$^{-1}$]',
    'v0r': r'$\Delta$ Centroid [m s$^{-1}$]',
    'v0z': r'$\Delta$ Centroid [m s$^{-1}$]',
    'vr_leftover': r'$\Delta$ Centroid [m s$^{-1}$]',    
    'peakintensity': r'$\Delta$ Peak Int. [K]'
}

clabel_ = clabels[args.moment].split(' [')
clabel_unit = unit #clabel_[1].split(']')[0]
clabel_prop = clabel_[0]#.split('$')[-1]
cu = (clabel_prop, clabel_unit)

clabels_grad = {
    'r': r'$\delta/\delta R$ (%s) [%s/au]'%cu,
    #'phi': r'$\delta/R\delta \phi$ (%s) [%s/au]'%cu,
    'phi': r'$\nabla_\phi$ (%s) [%s/au]'%cu,
    'peak': r'$\delta/\delta_{\rm peak}$ (%s) [%s/au]'%cu,
}

#***********
#MAKE PLOTS
#***********
if args.gradient=='peak':
    fig, ax = plt.subplots(ncols=1, nrows=2, figsize=(12, 6.5))
else:
    fig, ax = plt.subplots(ncols=1, nrows=3, figsize=(12, 10))

#Positive gradient    
_, _, cbar = make_polar_map(ufac*residuals, ufac*levels_grad,
                            R[args.surface]*u.m, phi[args.surface]*u.rad, Rout_plt,
                            fig=fig, ax=ax[0],
                            Rin=Rin_plt, gradient=args.gradient,
                            kwargs_gradient_peaks={'threshold': args.threshold},
                            findpeaks='pos', filepeaks='phi_gradient_peaks_%s_positive.txt'%mtags['base'],
                            cmap=cmap_grad, fmt=dfmt, clabel=clabels_grad[args.gradient])
if args.gradient!='peak':
    #Negative gradient
    _, _, cbar = make_polar_map(ufac*residuals, ufac*np.sort(-levels_grad),
                                R[args.surface]*u.m, phi[args.surface]*u.rad, Rout_plt,
                                fig=fig, ax=ax[1],
                                Rin=Rin_plt, gradient=args.gradient,
                                kwargs_gradient_peaks={'threshold': args.threshold},                                
                                findpeaks='neg', filepeaks='phi_gradient_peaks_%s_negative.txt'%mtags['base'],
                                cmap=cmap_grad_r, fmt=dfmt, clabel=clabels_grad[args.gradient])

#REFERENCE MAP
_, _, cbar = make_polar_map(ufac*residuals, ufac*levels_resid,
                            R[args.surface]*u.m, phi[args.surface]*u.rad, Rout_plt,
                            fig=fig, ax=ax[-1],
                            Rin=Rin_plt, gradient=0,
                            cmap=cmap_res, fmt='%4d', clabel=clabels[args.moment])


if args.moment in ['velocity', 'linewidth']:
    try:
        peaks_pos = np.loadtxt('phi_gradient_peaks_%s_positive.txt'%mtags['base']).T    
        peaks_neg = np.loadtxt('phi_gradient_peaks_%s_negative.txt'%mtags['base']).T

        kwargs_peaks = dict(color='indigo', lw=4, alpha=1.0, s=250)    
        ax[-1].scatter(peaks_pos[0], peaks_pos[1], marker='+', **kwargs_peaks)
        ax[-1].scatter(peaks_neg[0], peaks_neg[1], marker='_', **kwargs_peaks, label='Negative peak gradient')
        ax[0].scatter(None, None, label='Positive peak gradient', marker='+', **kwargs_peaks)
        ax[0].scatter(None, None, label='Negative peak gradient', marker='_', **kwargs_peaks)        
            
    except FileNotFoundError:
        pass

if args.show_legend:
    make_1d_legend(ax[0], handlelength=1.5, loc='lower center', bbox_to_anchor=(0.5, 1.2))
    
for axi in ax:
    if axi!=ax[-1]:
        axi.set_xlabel(None)
        kc = {'color': '1.0'}
    else:
        kc = {}

    if axi==ax[-1]:
        label_rings = True
    elif axi==ax[-2]:
        axi.set_ylabel(None)        
        label_rings = True
    else:
        axi.set_ylabel(None)
        label_rings = False

    make_substructures(axi, gaps=gaps, rings=rings, twodim=True, polar=True, label_rings=label_rings, kwargs_gaps=kc, kwargs_rings=kc)
    
mark_planet_location(ax[-1], args, edgecolors='k', lw=3.5, s=550, coords='disc', zfunc=z_func, zpars=z_pars, incl=incl, PA=PA, xc=xc, yc=yc, dpc=dpc)    
ax[0].set_title(ctitle, fontsize=16, color='k')

plt.savefig('gradient_residuals_deproj_%s_d%s_polar.png'%(mtags['base'], args.gradient), bbox_inches='tight', dpi=200)
show_output(args)



"""
Useful repeating code
"""
import json
import decimal
import warnings
import copy

from discminer.plottools import get_discminer_cmap
from discminer.core import Data
from discminer.rail import Rail
from discminer.disc2d import Model
import discminer.cart as cart

from astropy.io import fits
import astropy.units as u

import numpy as np
import matplotlib

disctex = {
    'dmtau': r'$\rm DM\,Tau$',
    'aatau': r'$\rm AA\,Tau$',
    'lkca15': r'$\rm LkCa15$',
    'hd34282': r'$\rm HD\,34282$',
    'mwc758': r'$\rm MWC\,758$', 
    'cqtau': r'$\rm CQ\,Tau$',
    'sycha': r'$\rm SY\,Cha$',
    'pds66': r'$\rm PDS\,66$',
    'hd135344': r'$\rm HD\,135344B$',
    'j1604': r'$\rm J1604$',
    'j1615': r'$\rm J1615$',
    'v4046': r'$\rm V4046\,Sgr$',
    'j1842': r'$\rm J1842$',
    'j1852': r'$\rm J1852$'
}

moltex = {
    '12co': r'$^{\rm 12}$CO',
    '13co': r'$^{\rm 13}$CO',
    'cs': r'CS'
}

warnings.filterwarnings("ignore", category=UserWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning)

SMALL_SIZE = 10
MEDIUM_SIZE = 15

    
def add_parser_args(parser,
                    moment=False, kernel=False, kind=False, surface=False, projection=False,
                    mask_minor=False, mask_major=False, Rinner=False, Router=False,
                    fold=False, writetxt=False, sigma=False, gradient=False, smooth=False
):

    def set_default(val, default):
        if isinstance(val, bool):
            return default
        else:
            return val
        
    if moment:
        d0 = set_default(moment, 'velocity')
        parser.add_argument('-m', '--moment', default=d0, type=str,
                            choices=['velocity', 'linewidth', 'lineslope', 'peakint', 'peakintensity'],
                            help="Type of moment map to be analysed. DEFAULTS to '%s'"%d0)
    if mask_minor:
        d0 = set_default(mask_minor, 30.0)
        parser.add_argument('-b', '--mask_minor', default=d0, type=float,
                            help="+- azimuthal mask around disc minor axis for calculation of vphi and vz velocity components. DEFAULTS to %.1f deg"%d0)
    if mask_major:
        d0 = set_default(mask_major, 30.0)
        parser.add_argument('-a', '--mask_major', default=d0, type=float,
                            help="+- azimuthal mask around disc major axis for calculation of vR velocity component. DEFAULTS to %.1f deg"%d0)
    if kernel:
        d0 = set_default(kernel, 'gaussian')
        parser.add_argument('-k', '--kernel', default=d0, type=str,
                            choices=['gauss', 'gaussian', 'bell', 'quadratic', 'dgauss', 'doublegaussian', 'dbell', 'doublebell'],
                            help="Kernel utilised for line profile fit and computation of moment maps. DEFAULTS to '%s'"%d0)
    if kind:
        d0 = set_default(kind, 'mask')
        parser.add_argument('-ki', '--kind', default=d0, type=str, choices=['mask', 'sum'],
                            help="How the upper and lower surface kernel profiles must be merged. DEFAULTS to '%s'"%d0)
    if surface:
        d0 = set_default(surface, 'upper')
        parser.add_argument('-s', '--surface', default=d0, type=str,
                            choices=['up', 'upper', 'low', 'lower'],
                            help="Use upper or lower surface moment map. DEFAULTS to '%s'"%d0)
    if fold:
        d0 = set_default(fold, 'absolute')        
        parser.add_argument('-f', '--fold', default=d0, type=str,
                            choices=['absolute', 'standard'],
                            help="if moment=velocity, fold absolute or standard velocity residuals. DEFAULTS to '%s'"%d0)
    if projection:
        d0 = set_default(fold, 'cartesian')                
        parser.add_argument('-p', '--projection', default=d0, type=str,
                            choices=['cartesian', 'polar'],
                            help="Project residuals onto a cartesian or a polar map. DEFAULTS to '%s'"%d0)
    if writetxt:
        d0 = set_default(writetxt, 1)                
        parser.add_argument('-w', '--writetxt', default=d0, type=int,
                            choices=[0, 1],
                            help="write output into txt file(s). DEFAULTS to %d"%d0)
    if Rinner:
        d0 = set_default(Rinner, 1.0)                        
        parser.add_argument('-i', '--Rinner', default=d0, type=float,
                            help="Number of beams to mask out from disc inner region. DEFAULTS to %.2f"%d0)
    if Router:
        d0 = set_default(Router, 0.98)                
        parser.add_argument('-o', '--Router', default=d0, type=float,
                            help="Fraction of Rout to consider as the disc outer radius for the analysis. DEFAULTS to %.2f"%d0)
    if sigma:
        d0 = set_default(sigma, 5)                
        parser.add_argument('-si', '--sigma', default=d0, type=float,
                            help="Mask out pixels with values below sigma threshold. DEFAULTS to %.1f"%d0)
    if gradient:
        d0 = set_default(gradient, 'r')
        parser.add_argument('-g', '--gradient', default=d0, type=str, choices=['peak', 'r', 'phi'],
                            help="Coordinate along which the gradient will be computed. If 'peak', the maximum gradient is computed. DEFAULTS to '%s'"%d0)

    if smooth:
        d0 = set_default(smooth, 0.0)
        parser.add_argument('-sm', '--smooth', default=d0, type=float,
                            help="Smooth moment_data using skimage.restoration library. DEFAULTS to %.1f"%d0)        

    args = parser.parse_args()

    try:
        if args.moment=='peakint':
            args.moment = 'peakintensity'
    except AttributeError:
        pass

    try:
        if args.surface=='up':
            args.surface = 'upper'
    except AttributeError:
        args.surface = 'upper'

    try:
        if args.surface=='low':
            args.surface = 'lower'
    except AttributeError:
        args.surface = 'upper'

    try:
        if args.kernel=='gauss':
            args.kernel = 'gaussian'
    except AttributeError:
        args.kernel = 'gaussian'
    
    try:
        if args.kernel=='dgauss':
            args.kernel = 'doublegaussian'
    except AttributeError:
        args.kernel = 'gaussian'

    try:
        if args.kernel=='dbell':
            args.kernel = 'doublebell'
    except AttributeError:
        args.kernel = 'gaussian'
    
    return args


def init_data_and_model(parfile='parfile.json', Rmin=0, Rmax=1.1, init_model=True):
    #Rmin: If dimensionless, fraction of beam_size
    #Rmax: If dimensionless, fraction of Rout    
    with open(parfile) as jf:
        pars = json.load(jf)

    meta = pars['metadata']
    best = pars['best_fit']
    custom = pars['custom']

    file_data = meta['file_data']
    Rout = best['intensity']['Rout']
    dpc = meta['dpc']*u.pc

    datacube = Data(file_data, dpc) # Read data and convert to Cube object

    if init_model:
        Rmax = Rmax*Rout*u.au
        model = Model(datacube, Rmax=Rmax, Rmin=Rmin, prototype=True)
        
        model.velocity_func = model.keplerian_vertical # vrot = sqrt(GM/r**3)*R
        model.line_profile = model.line_profile_bell

        if 'sum' in meta['kind']:
            model.line_uplow = model.line_uplow_sum
        else:
            model.line_uplow = model.line_uplow_mask

        if 'I2pwl' in meta['kind']:
            model.intensity_func = cart.intensity_powerlaw_rbreak
        elif 'I2pwlnosurf' in meta['kind']:
            model.intensity_func = cart.intensity_powerlaw_rbreak_nosurf    
        else:
            model.intensity_func = cart.intensity_powerlaw_rout

        if 'surf2pwl' in meta['kind']:
            model.z_upper_func = cart.z_upper_powerlaw
            model.z_lower_func = cart.z_lower_powerlaw
        else:
            model.z_upper_func = cart.z_upper_exp_tapered
            model.z_lower_func = cart.z_lower_exp_tapered
            
        #****************
        #PROTOTYPE PARAMS
        #****************
        model.params = copy.copy(best)
        model.params['intensity']['I0'] /= meta['downsamp_factor']

        return datacube, model

    else:
        return datacube

    
def read_json(decimals=True): #Read json file again but keep track of (if) decimals
    if decimals:
        parse_func = lambda x: decimal.Decimal(str(x))
    else:
        parse_func = float

    with open('parfile.json') as json_file: 
        pars = json.load(json_file, parse_float=parse_func)
    return pars
            
def get_2d_plot_decorators(moment, unit_simple=False, fmt_vertical=False):    
    I_res2abs = 3.0
    v_res2abs = 20.0
    L_res2abs = 5.0
    Ls_res2abs = 3.0

    pars = read_json()
    vsys = float(pars['best_fit']['velocity']['vsys'])
    custom = pars['custom']
    meta = pars['metadata']

    try:
        ctitle = disctex[meta['disc']]+' '+moltex[meta['mol']]
        
    except KeyError:
        ctitle = meta['disc']+' '+meta['mol']
        
    if moment=='velocity':
        clim = custom['vlim']
        fclim = float(clim)        
        unit = '[km s$^{-1}$]'
        if unit_simple: unit = 'km/s'
        clabel = r'Centroid Velocity %s'%unit
        cmap_mom = cmap_res = get_discminer_cmap('velocity')
        levels_im = np.linspace(-v_res2abs*fclim, v_res2abs*fclim, 64)+vsys
        levels_cc = np.linspace(-2, 2, 9)+vsys
    
    if moment=='linewidth':
        clim = custom['Llim']
        fclim = float(clim)
        unit = '[km s$^{-1}$]'
        if unit_simple: unit = 'km/s'        
        clabel = r'Line width %s'%unit
        cmap_mom = get_discminer_cmap('linewidth')
        cmap_res = get_discminer_cmap('linewidth', kind='residuals')
        levels_im = np.linspace(0.0, L_res2abs*fclim, 48)
        levels_cc = np.linspace(0.2, 0.8, 4)

    if moment=='lineslope':
        clim = custom['Lslim']
        fclim = float(clim)        
        unit = ''
        clabel = r'Line slope'
        cmap_mom = get_discminer_cmap('linewidth')
        cmap_res = get_discminer_cmap('linewidth', kind='residuals')
        levels_im = np.linspace(0.0, Ls_res2abs*fclim, 48)
        levels_cc = np.linspace(1.0, Ls_res2abs*fclim, 5)
    
    if moment in ['peakintensity', 'peakint']:
        clim = custom['Ilim']
        fclim = float(clim)        
        unit = '[K]'
        if unit_simple: unit = 'K'        
        clabel = r'Peak Intensity %s'%unit
        cmap_mom = get_discminer_cmap('intensity')    
        cmap_res = get_discminer_cmap('intensity', kind='residuals')
        levels_im = np.linspace(0.0, I_res2abs*fclim, 48)
        levels_cc = np.linspace(10, 50, 5)

    sp_dec = str(clim).split('.')
    nsign = 0
    if fmt_vertical:
        nsign = 1
    
    if len(sp_dec)==2:
        ndigs = len(sp_dec[0]) + len(sp_dec[-1])        
        cfmt = '%'+'%d.%df'%(ndigs+1+nsign, len(sp_dec[-1]))
    else:
        ndigs = len(sp_dec[0])
        cfmt = '%'+'%d.1f'%(ndigs+1+1+nsign) #ndigs + . + 1f + nsign
    
    return ctitle, clabel, fclim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit

def get_1d_plot_decorators(moment, tag=''):
    I_res2abs = 5.0
    v_res2abs = 20.0
    L_res2abs = 7.0
    Ls_res2abs = 3.0
        
    pars = read_json(decimals=False)
    vsys = float(pars['best_fit']['velocity']['vsys'])
    custom = pars['custom']
    meta = pars['metadata']

    if len(tag)>0: tag = tag+' '
        
    if moment=='velocity':
        unit = '[km s$^{-1}$]'
        clim0_res = -custom['vlim']
        clim1_res = custom['vlim']    
        clim0, clim1 = 0.0, v_res2abs*clim1_res + vsys
        clabel = clabel_res = None #Modified within each script
        
    if moment=='linewidth':
        clim0_res = -custom['Llim']
        clim1_res = custom['Llim']
        clim0, clim1 = 0, L_res2abs*clim1_res
        unit = '[km s$^{-1}$]'
        clabel = r'Line width %s%s'%(tag, unit)
        clabel_res = r'L. width residuals %s'%unit    

    if moment=='lineslope':
        clim0_res = -custom['Lslim']
        clim1_res = custom['Lslim']
        clim0, clim1 = 0, Ls_res2abs*clim1_res
        unit = ''
        clabel = r'Line slope'
        clabel_res = r'L. slope residuals'
        
    if moment in ['peakintensity', 'peakint']:
        clim0_res = -custom['Ilim']
        clim1_res = custom['Ilim']
        clim0, clim1 = 0.0, I_res2abs*clim1_res
        unit = '[K]'
        clabel = r'Peak Intensity %s%s'%(tag, unit)
        clabel_res = r'Peak Int. residuals %s'%unit    
    
    return clabel, clabel_res, clim0, clim0_res, clim1, clim1_res, unit

def get_noise_mask(datacube, thres=4, return_mean=True):
    #std from line-free channels, per pixel
    data = datacube.data
    noise = np.std( np.append(data[:5,:,:], data[-5:,:,:], axis=0), axis=0) 
    if return_mean:
        noise_mean = np.mean(noise)
        mask = np.nanmax(data, axis=0) < thres*noise_mean
        return noise_mean, mask
    else:
        mask = np.nanmax(data, axis=0) < thres*noise
        return noise, mask

def load_moments(
        args, moment=None, kernel=None, mask=[],
        clip_Rgrid=None, clip_Rmin=0*u.au, clip_Rmax=np.inf*u.au
):    
    if moment is None:
        moment = args.moment

    if kernel is None:
        kernel = args.kernel
        
    if kernel in ['gauss', 'gaussian', 'quadratic', 'bell']:
        tag_surf = 'both'
        ref_surf = 'upper' #For contour plots
    else:
        if args.surface in ['up', 'upper']:
            tag_surf = 'up'
            ref_surf = 'upper'
        elif args.surface in ['low', 'lower']:
            tag_surf = 'low'
            ref_surf = 'lower'
            
    if kernel in ['gauss', 'gaussian']:
        tag_base = f'{moment}_gaussian'
    elif kernel in ['quadratic']:
        tag_base = f'{moment}_quadratic'        
    elif kernel in ['bell']:
        tag_base = f'{moment}_bell'
    elif kernel in ['dgauss', 'doublegaussian']:
        tag_base = f'{moment}_{tag_surf}_doublegaussian_{args.kind}'
    elif kernel in ['dbell', 'doublebell']:
        tag_base = f'{moment}_{tag_surf}_doublebell_{args.kind}'

    #Read and mask moment maps, and compute residuals
    moment_data = fits.getdata(tag_base+'_data.fits')
    moment_model = fits.getdata(tag_base+'_model.fits')

    moment_data[mask] = np.nan
    moment_model[mask] = np.nan

    try:
        if args.smooth>0.0:
            from scipy.ndimage import gaussian_filter 
            moment_data = gaussian_filter(moment_data, args.smooth)
            moment_model = gaussian_filter(moment_model, args.smooth)                        
    except AttributeError:
        pass
    
    def clip_radially(prop2d): #should be within masking function in discminer         
        Rmin = clip_Rmin.to('au').value
        Rgrid = np.nan_to_num(clip_Rgrid).to('au').value
        Rmax = clip_Rmax.to('au').value
        return np.where((Rgrid<Rmin) | (Rgrid>Rmax), np.nan, prop2d)    

    if isinstance(clip_Rgrid, u.Quantity):
        moment_data = clip_radially(moment_data)
        moment_model = clip_radially(moment_model)

    residuals = moment_data - moment_model

    if args.surface in ['low', 'lower']:
        pass
        #from scipy.ndimage import gaussian_filter
        #moment_data = gaussian_filter(moment_data, 0.5)
        #moment_data[moment_data > 50.0] = np.nan

    return moment_data, moment_model, residuals, dict(surf=tag_surf, ref_surf=ref_surf, base=tag_base)
    
    
def load_disc_grid():
    R = dict(
        upper=np.load('upper_R.npy'),
        lower=np.load('lower_R.npy')
    )

    phi = dict(
        upper=np.load('upper_phi.npy'),
        lower=np.load('lower_phi.npy')
    )

    z = dict(
        upper=np.load('upper_z.npy'),
        lower=np.load('lower_z.npy')
    )
    return R, phi, z

def make_1d_legend(ax, **kwargs):
    kwargs_def = dict(
        frameon=False,
        fontsize=MEDIUM_SIZE,
        ncol=3,
        handlelength=2.0,
        handletextpad=0.5,
        borderpad=0.0,
        columnspacing=1.5,
        loc='lower right',
        bbox_to_anchor=(1.0, 1.0)
    )
    kwargs_def.update(kwargs)
    return ax.legend(**kwargs_def)

def make_and_save_filaments(model, map2d,
                            writefits=True, tag='',
                            return_all=False, cmap='jet',
                            surface='upper'
):
    mrail = Rail(model, map2d)
    Rgrid = model.projected_coords['R'][surface]
    
    fil_pos, fil_neg = mrail.make_filaments(
        smooth_size=0.1*model.beam_size,
        adapt_thresh=0.5*model.beam_size,
        size_thresh=100*u.pix**2
    )

    """
    masked_pos = np.ma.masked_where(fil_pos.skeleton!=1, fil_pos.skeleton)
    masked_neg = np.ma.masked_where(fil_neg.skeleton!=1, fil_neg.skeleton)    
    ax.imshow(masked_pos, cmap='Reds_r', alpha=1, zorder=3, extent = [mm.X.min(),mm.X.max(), mm.Y.min(),mm.Y.max()], origin='lower')
    ax.imshow(masked_neg, cmap='Blues_r', alpha=1, zorder=3, extent = [mm.X.min(),mm.X.max(), mm.Y.min(),mm.Y.max()], origin='lower')
    """

    fil_pos_list = [fil_pos.skeleton, fil_pos.skeleton_longpath]
    fil_neg_list = [fil_neg.skeleton, fil_neg.skeleton_longpath]

    check_outer_nans = lambda fil: np.isnan(Rgrid[fil.pixel_coords]).any()
        
    for i,fil in enumerate(fil_pos.filaments):
        
        if check_outer_nans(fil):
            del fil_pos.filaments[i]
            continue
        
        fil_i = np.zeros_like(fil_pos.skeleton)
        fil_i[fil.pixel_coords] = 1 
        fil_pos_list.append(fil_i)

        
    for i,fil in enumerate(fil_neg.filaments):
        
        if check_outer_nans(fil):
            del fil_neg.filaments[i]
            continue
        
        fil_i = np.zeros_like(fil_neg.skeleton)
        fil_i[fil.pixel_coords] = 1 
        fil_neg_list.append(fil_i)

    if writefits:
        if len(tag)>0:
            if tag[0]!='_':
                tag='_'+tag
                
        fits.writeto('filaments_pos%s.fits'%tag, np.asarray(fil_pos_list), header=model.header, overwrite=True)
        fits.writeto('filaments_neg%s.fits'%tag, np.asarray(fil_neg_list), overwrite=True)

    cmap = matplotlib.cm.get_cmap(cmap)
    npos = len(fil_pos.filaments)
    nneg = len(fil_neg.filaments)
    cpos = np.linspace(0.6, 1.0, npos)        
    cneg = np.linspace(0.4, 0.0, nneg)
    colors_dict = {}
    colors_dict.update({i+1: matplotlib.colors.to_hex(cmap(cpos[i])) for i in range(npos)})
    colors_dict.update({-i-1: matplotlib.colors.to_hex(cmap(cneg[i])) for i in range(nneg)})
        
    if return_all:
        return fil_pos, fil_neg, fil_pos_list, fil_neg_list, colors_dict       

    else:
        return fil_pos_list, fil_neg_list, colors_dict




from discminer.mining_control import _mining_pick
from discminer.mining_utils import (init_data_and_model,
                                    get_noise_mask,
                                    get_2d_plot_decorators,
                                    load_moments,
                                    load_disc_grid,
                                    mark_planet_location,
                                    make_masks,
                                    show_output)
from discminer.pick import Pick
from discminer.grid import GridTools

from discminer.plottools import (get_discminer_cmap,
                                 append_sigma_panel,
                                 mod_minor_ticks,
                                 mod_major_ticks,
                                 use_discminer_style,
                                 make_substructures,
                                 make_round_map,
                                 make_polar_map,
                                 get_cmap_from_color,
                                 make_clusters_1d,
                                 make_1d_legend)

import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u
from astropy.io import fits

import json
import os

use_discminer_style()

if __name__ == '__main__':
    parser = _mining_pick(None)
    args = parser.parse_args()

#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

vel_sign = best['velocity']['vel_sign']
vsys = best['velocity']['vsys']
Rout = best['intensity']['Rout']
incl = best['orientation']['incl']
PA = best['orientation']['PA']
xc = best['orientation']['xc']
yc = best['orientation']['yc']

gaps = custom['gaps']
rings = custom['rings']

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment, unit_simple=True, fmt_vertical=True)

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc

#*******************
#LOAD DATA AND MODEL
#*******************
datacube, model = init_data_and_model()
vchannels = datacube.vchannels
model.make_model()

beam_au = datacube.beam_size.to('au').value
if args.absolute_Rinner>=0:
    Rmod_in = args.absolute_Rinner
else:
    Rmod_in = args.Rinner*beam_au

if args.absolute_Router>=0:
    Rmod_out = args.absolute_Router
else:
    Rmod_out = args.Router*Rout

Rmax = 1.1*Rmod_out*u.au #Max window extent, 10% larger than disc Rout

R, phi, z = load_disc_grid()
noise_mean, mask = get_noise_mask(datacube, thres=2,
                                  mask_phi={'map2d': np.degrees(phi['upper']),
                                            'lims': args.mask_phi},
                                  mask_R={'map2d': R['upper']/au_to_m,
                                          'lims': args.mask_R}
)

#noise_mean, mask = get_noise_mask(datacube, thres=2)

if args.surface in ['up', 'upper']:
    z_func = model.z_upper_func
    z_pars = best['height_upper']

elif args.surface in ['low', 'lower']:
    z_func = model.z_lower_func
    z_pars = best['height_lower']

#*************************
#LOAD MOMENT MAPS
moment_data, moment_model, residuals, mtags = load_moments(args, mask=mask)
    
if args.moment=='velocity' and args.fold=='absolute':
    residuals = np.abs(moment_data-vsys) - np.abs(moment_model-vsys)
    residuals[mask] = np.nan
    if args.percentage_kepler:
        ref_surf = mtags['ref_surf']
        coords = {'R': model.projected_coords['R']['upper']}
        velocity_kepler = model.get_attribute_map(coords, 'velocity', surface=ref_surf) * vel_sign    
        residuals = residuals/velocity_kepler

#********************
def get_sky_coords(rp, phip, midplane=True):
    phii = np.radians(phip)
    zp = z_func({'R': rp*u.au.to('m')}, **z_pars)*u.m.to('au')
    if midplane:
        zp *= 0
    xi,yi,zi = GridTools.get_sky_from_disc_coords(rp, phii, zp, incl, PA, xc, yc)
    PAsky = np.arctan2(yi, xi) - np.pi/2
    rsky = np.hypot(xi-xc, yi-yc)/dpc.value
    return rsky, np.degrees(PAsky)

#*******************
#FIND PEAK RESIDUALS
#*******************
R_prof = np.arange(Rmod_in, Rmod_out, beam_au/5)
xlim0, xlim1 = 0.5*R_prof[0], 1.05*R_prof[-1]

fig, ax = plt.subplots(ncols=2, nrows=1, figsize=(11,5))    
ax_c = fig.add_axes([0.8,0.65,0.23*5/11,0.23]) #AxesSubplot for showing contour colours 

pick = Pick(model, residuals, R_prof, fold=True, color_bounds = np.array([0.33, 0.66, 1.0])*Rmod_out, ax2=ax_c)
xf, yf, folded_orig, folded_map = pick.make_2d_map(return_coords=True) #Map where peaks will be picked from
        
figh, axh = plt.subplots(ncols=1, nrows=1, figsize=(9,6))    
pick.find_peaks(clean_thres=args.clean_thres, fig_ax_histogram=(figh, axh), clean_histogram=True)
figh.savefig('histogram_peak_residuals_%s.png'%mtags['base'], bbox_inches='tight', dpi=200)

lev = pick.lev_list
color = pick.color_list
peak_resid = pick.peak_resid
peak_angle = pick.peak_angle

model.make_emission_surface(ax_c)
model.make_disc_axes(ax_c)
ax_c.axis('off')
    
#*******************
#SHOW PEAK RESIDUALS
#*******************
kwargs_sc = dict(facecolors=color, edgecolors='0.1', lw=0.7, alpha=0.7, zorder=10)
color_global_peak = 'lime'
color_cluster_peak = 'lime'

ax[0].set_xlabel(r'Azimuth [deg]')
ax[0].set_ylabel('Peak residual [%s]'%unit)
ax[1].set_xlabel('Radius [au]')

for axi in ax: 
    axi.tick_params(labelbottom=True, top=True, right=True, which='both', direction='in')    
    mod_major_ticks(axi, nbins=8)
    mod_minor_ticks(axi)
ax[0].set_xlim(-95,95)
ax[0].set_xticks(np.arange(-90,90+1,30))    
ax[1].tick_params(labelleft=False)

ax[0].scatter(peak_angle, peak_resid, **kwargs_sc)
ax[1].scatter(lev, peak_resid, **kwargs_sc)

ax[0].axvline(pick.peak_global_angle, lw=3.5, c=color_global_peak, label='global peak', zorder=0)
ax[1].axvline(pick.peak_global_radius, lw=3.5, c=color_global_peak, zorder=0) 
ax[0].legend(frameon=False, fontsize=15, handlelength=1.0, loc='lower left', bbox_to_anchor=(-0.04, 0.98))

make_substructures(ax[1], gaps=gaps, rings=rings)
append_sigma_panel(fig, ax, peak_resid, weights=pick.peak_weight, hist=True)
fig.savefig('peak_residuals_%s.png'%mtags['base'], bbox_inches='tight', dpi=200)
#plt.show()
plt.close()

#sys.exit()

#*************
#FIND CLUSTERS
#*************
if args.clusters:
    try:
        pick.find_clusters(n_phi=args.nphi, n_R=args.nr)
        found_clusters = True    
    except np.linalg.LinAlgError as e:
        print(30*'*')    
        print(repr(e))
        print('Cluster finder algorithm failed due to the low number of input points. Try changing the number of clusters to use or the disc extent...')
        print(30*'*')        
        found_clusters = False
else:
    found_clusters = False    
    
#*************
#PLOT CLUSTERS
#*************
if found_clusters and args.clusters:
    fig, ax = make_clusters_1d(pick, which='phi')
    plt.savefig('clusters_phi_peak_residuals_%s_%dclust.png'%(mtags['base'], args.nphi), bbox_inches='tight', dpi=200)
    plt.close()
    #plt.show()

    fig, ax = make_clusters_1d(pick, which='r')
    plt.savefig('clusters_r_peak_residuals_%s_%dclust.png'%(mtags['base'], args.nr), bbox_inches='tight', dpi=200)
    plt.close()
    #plt.show()

#***************
#SHOW FOLDED MAP
#***************
R = model.projected_coords['R']
phi = model.projected_coords['phi']

Xproj = R[args.surface]*np.cos(phi[args.surface])
Yproj = R[args.surface]*np.sin(phi[args.surface])

cos_peak = np.cos(np.radians(peak_angle))
sin_peak = np.sin(np.radians(peak_angle))

if args.projection=='cartesian':
    levels_resid = np.linspace(-clim, clim, 32)
        
    fig, ax = make_round_map(folded_map, levels_resid, pick.X*u.au, pick.Y*u.au, R_prof[-1]*u.au,
                             z_func=z_func, z_pars=z_pars, incl=incl, PA=PA, xc=xc, yc=yc,
                             cmap=cmap_res, clabel=unit, fmt=cfmt,
                             make_cbar=args.colorbar, 
                             rings=rings,                             
                             mask_wedge=(90, 270)*u.deg,
                             mask_inner=R_prof[0]*u.au)
    make_substructures(
        ax, gaps=gaps, rings=rings, twodim=True, label_rings=True,
    )#kwargs_rings={'lw': 0.6})

    if args.show_peaks:
        ax.scatter(lev*cos_peak, lev*sin_peak, edgecolors='none', facecolors=color, alpha=0.2, s=100+150, zorder=20)
        ax.scatter(lev*cos_peak, lev*sin_peak, edgecolors='none', facecolors=color, alpha=1.0, s=10+30, zorder=20)
        ax.scatter(lev*cos_peak, lev*sin_peak, edgecolors='0.3', facecolors='none', alpha=1.0, s=100+150, zorder=20)

    #Show global peak
    sb = 450
    kwargs_global = dict(edgecolors='k', facecolors=color_global_peak, marker='o', lw=2.5, alpha=0.7)
    
    if args.show_global:
        ax.scatter(pick.peak_global_radius*np.cos(np.radians(pick.peak_global_angle)),
                   pick.peak_global_radius*np.sin(np.radians(pick.peak_global_angle)),
                   s=sb+250, zorder=21, **kwargs_global)
               
        ax.scatter(None, None, label='Global peak', s=sb, **kwargs_global) #for legend
    
    print ('Global peak residual found at R=%.1f au, phi=%.1f deg, in disc frame coordinates...'%(pick.peak_global_radius, pick.peak_global_angle))
    print ('Global peak in sky coords, R=%.3f arcsec, PA=%.1f deg...'%get_sky_coords(pick.peak_global_radius, pick.peak_global_angle))
    
    if found_clusters and args.clusters:
        
        if len(pick.acc_peaks_phi)>0 and len(pick.acc_peaks_R)>0:
            #Show cluster peak
            kwargs_cluster = dict(edgecolors='k', facecolors=color_cluster_peak, marker='X', lw=2.5, alpha=0.7)
            ax.scatter(pick.acc_R*np.cos(np.radians(pick.acc_phi)),
                       pick.acc_R*np.sin(np.radians(pick.acc_phi)),                       
                       s=sb+250, zorder=21, **kwargs_cluster)

            ax.scatter(None, None, label='Cluster peak', s=sb, **kwargs_cluster) #for legend
            
            print ('Weighted centre of accepted clusters found at R=%.1f au, phi=%.1f deg, in disc frame coordinates...'%(pick.acc_R, pick.acc_phi))
            print ('Cluster in sky coords, R=%.3f arcsec, PA=%.1f deg...'%get_sky_coords(pick.acc_R, pick.acc_phi))
            
    #Mark planet location if passed as an arg
    kwargs_planet = dict(edgecolors='gold', facecolors='none', marker='o', lw=4.5, alpha=1.0, zorder=22)    

    mark_planet_location(ax, args, edgecolors='k', lw=3.5, s=sb+200, coords='disc', zfunc=z_func, zpars=z_pars, incl=incl, PA=PA, xc=xc, yc=yc, dpc=dpc)    
    
    #mark_planet_location(ax, args, s=sb+200, **kwargs_planet)
    ax.scatter(None, None, label='Planet', s=sb, **kwargs_planet) #for legend
            
    if len(args.rp)>0 and args.show_legend:
        make_1d_legend(ax, handlelength=1.5, loc='lower center', bbox_to_anchor=(0.5, 1.02))
        
    ax.set_title('%s, folded map'%ctitle, fontsize=16, color='k')

    
elif args.projection=='polar': #Currently displaying folded map only
    levels_resid = np.linspace(-clim, clim, 48)

    Rf = np.hypot(pick.X, pick.Y)
    phif = np.arctan2(pick.Y, pick.X)
    
    fig, ax, cbar = make_polar_map(folded_map, levels_resid,
                                   Rf*u.au, phif*u.rad, R_prof[-1]*u.au,
                                   cmap=cmap_res, clabel=unit, fmt=cfmt)
    ax.fill_between([-180, -90], R_prof[-1], color='k', alpha=0.5)
    ax.fill_between([90, 180], R_prof[-1], color='k', alpha=0.5)
    
    make_substructures(
        ax, gaps=gaps, rings=rings, twodim=True, label_rings=True, polar=True
    )

    
if len(args.mask_R)>0 or len(args.mask_phi)>0:
    make_masks(ax, args.mask_R, args.mask_phi, Rmax=Rmod_out)

    
symm_mean = np.nanmean(peak_resid)
symm_std = np.nanstd(peak_resid)
symm = symm_mean + symm_std
print ('Non-axisymmetric factor from peak folded residuals mean, std, (mean+1std): %.3f, %.3f, %.3f km/s'%(symm_mean, symm_std, symm))
    
#*********************
#MAKE FILES AND PLOTS
#*********************
#Write Pick summary file
if args.writetxt:    
    pick.writetxt(filename='pick_summary_%s.txt'%mtags['base'])

#Write interpolated folded_map into file
if args.writefits:
    fits.writeto(os.path.join(mtags['dir_model'], mtags['base']+'_foldedresiduals.fits'), folded_map, header=datacube.header, overwrite=True)    

#Write folded values into file, along with original x,y coordinates:
if args.writetxt:
    tmp = np.asarray([xf, yf, folded_orig]).T
    np.savetxt(os.path.join(mtags['dir_model'], mtags['base']+'_foldedresiduals.txt'), tmp, fmt='%.3f')
    
plt.savefig('folded_residuals_deproj_%s_%s.png'%(mtags['base'], args.projection), bbox_inches='tight', dpi=200)
show_output(args)



from discminer.core import Data
from discminer.cube import Cube
from discminer.disc2d import Model
from discminer.plottools import use_discminer_style
import discminer.cart as cart

from utils import add_parser_args, get_noise_mask, get_2d_plot_decorators, init_data_and_model

import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u

import json
import copy
from argparse import ArgumentParser

use_discminer_style()

parser = ArgumentParser(prog='plot channel maps', description='Plot channel maps and residuals')
parser.add_argument('-nc', '--nchans', default=7, type=int, help="Number of channels to be plotted")
args = add_parser_args(parser, sigma=3)

#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

chan_step = custom['chan_step']

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc

mol_tex = {
    '12co': r'$^{12}$CO',
    '13co': r'$^{13}$CO',
    'cs': r'CS'
}
#*******************
#LOAD DATA AND MODEL
#*******************
datacube, model = init_data_and_model(Rmin=0, Rmax=1.2)
vchannels = datacube.vchannels
pix_downsamp = model.grid['step']*meta['downsamp_fit']/au_to_m

#Useful definitions for plots
xmax = model.skygrid['xmax'] 

if meta['mol'] in ['13co', 'cs']:
    xlim = 0.7*xmax/au_to_m
else:
    xlim = xmax/au_to_m

extent= np.array([-xmax, xmax, -xmax, xmax])/au_to_m
  
#**************************
#MAKE MODEL (2D ATTRIBUTES)
#**************************
modelcube = model.make_model(make_convolve=True)
modelcube.convert_to_tb(writefits=False)
datacube.convert_to_tb(writefits=False)

#*****************
#PLOT CHANNEL MAPS
#*****************
# Import CMasher
cmap = None
if 'sum' in meta['kind']:
    import cmasher as cmr
    cmap = plt.get_cmap('cmr.rainforest_r')

nchans = args.nchans
idlim = int(0.5*chan_step*(nchans-1))
plot_channels = np.linspace(-idlim,idlim,nchans) + np.argmin(np.abs(vchannels-best['velocity']['vsys'])) 

kw_channels = dict(channels={'indices': plot_channels}, ncols=nchans, contours_from=model,
                   xlims = (-xlim, xlim), ylims = (-xlim, xlim), 
                   unit_coordinates='au', unit_intensity=None, projection=None)
                   

fig, ax = plt.subplots(nrows=3, ncols=nchans, figsize=(2.5*nchans, 2.5*3)) #Master figure
plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=0.05, hspace=0.05)

#DATA CHANNELS
fig, ax0, im0, cbar0 = datacube.make_channel_maps(fig=fig, ax=ax[0], cmap=cmap, **kw_channels)
fig, ax1, im1, cbar1 = modelcube.make_channel_maps(fig=fig, ax=ax[1], levels=im0[0].levels, annotate_channels=False, cmap=cmap, **kw_channels)


#RESIDUAL CHANNELS
noise_mean, mask = get_noise_mask(datacube)
residualscube = Cube(datacube.data-modelcube.data, datacube.header, datacube.vchannels, dpc, beam=datacube.beam)

fig, ax2, im2, cbar2 = residualscube.make_channel_maps(fig=fig, ax=ax[2],
                                                       kind='residuals',
                                                       annotate_channels=False,
                                                       mask_under=args.sigma*noise_mean,
                                                       levels=np.linspace(-custom['Ilim'], custom['Ilim'], 32),
                                                       #cmap=cmap_res,
                                                       **kw_channels)

cbar0.remove()

for axi in ax[:-1,0]:
    axi.set_xlabel('')    
    axi.tick_params(labelleft=False, labelbottom=False)

#ax[0,0].set_title(meta['disc'].upper(), pad=10, loc='left')
ax[0,0].set_ylabel(mol_tex[meta['mol']] + ' Data', color='k', fontsize=13, labelpad=6)
ax[1,0].set_ylabel('Model', color='k', fontsize=13, labelpad=4)


for axi in ax[:,int(round(0.5*(nchans+1)))-1]:
    model.make_emission_surface(
        axi,
        kwargs_R={'colors': '0.4', 'linewidths': 0.4},
        kwargs_phi={'colors': '0.4', 'linewidths': 0.3}
    )
    
    model.make_disc_axes(axi)

plt.savefig('channel_maps_residuals_%s_%s.png'%(meta['disc'], meta['mol']), bbox_inches = 'tight', dpi=200)    
plt.show()




from discminer.mining_control import _mining_spectra
from discminer.mining_utils import (get_noise_mask,
                                    get_2d_plot_decorators,
                                    init_data_and_model,
                                    load_moments,
                                    show_output)

from discminer.core import Data
from discminer.cube import Cube
from discminer.grid import GridTools
from discminer.tools import fit_kernel
from discminer.plottools import (make_up_ax,
                                 mod_minor_ticks,
                                 mod_major_ticks,
                                 mod_nticks_cbars,
                                 use_discminer_style)


import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from astropy import units as u
from astropy.io import fits

import json
import copy
from argparse import ArgumentParser

use_discminer_style()
matplotlib.rcParams['hatch.linewidth'] = 1.3

if __name__ == '__main__':
    parser = _mining_spectra(None)
    args = parser.parse_args()

#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']
Rout = best['intensity']['Rout']

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment)

"""
chot = '0.5'
cmap_mom.set_under(chot)
cmap_mom.set_over(chot)

cmap_res.set_under(chot)
cmap_res.set_over(chot)
"""

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc

mol_tex = {
    '12co': r'$^{12}$CO',
    '13co': r'$^{13}$CO',
    'cs': r'CS'
}

R_lev = np.arange(25, 0.98*Rout, 50)*au_to_m

#*******************
#LOAD DATA AND MODEL
#*******************
datacube, model = init_data_and_model(Rmin=0, Rmax=1.0)
noise_mean, mask = get_noise_mask(datacube, thres=0)

pix_downsamp = model.grid['step']*meta['downsamp_fit']/au_to_m
step = model.grid['step']/au_to_m

#Useful definitions for plots
xmax = model.skygrid['xmax'] 
xlim = 1.1*Rout
extent= np.array([-xmax, xmax, -xmax, xmax])/au_to_m
plot_title = ctitle + r' $-$ ' + clabel.split('[')[0]

#**************************
#MAKE MODEL (2D ATTRIBUTES)
#**************************
model.make_model()
modelcube = Data('cube_model_%s.fits'%tag, dpc)
#modelcube = model.make_model()
modelcube.convert_to_tb(writefits=False, planck=args.planck)
datacube.convert_to_tb(writefits=False, planck=args.planck)

#*************************
#LOAD MOMENT MAPS
moment_data, moment_model, residuals, mtags = load_moments(args, mask=mask)

#Show Data or Model?
if args.type=='data':
    map2d = moment_data
    cube = datacube
elif args.type=='model':
    map2d = moment_model
    cube = modelcube
    
vchannels = cube.vchannels

#Show Moment map or Channel map?
if args.channel_id > -1:
    map2d = cube.data[args.channel_id]
    map2d[mask] = np.nan
    plot_title = 'Channel: %.1f km/s'%vchannels[args.channel_id] + r' $-$ ' + clabel.split('[')[0]
    
#**************
#MAKE PLT AXES
#**************
figx = figy = 9

ncols = 5
nrows = 5

dw = 0.9/ncols
dh = dw*figx/figy

fig = plt.figure(figsize=(figx,figy))

ax = [
    [
        fig.add_axes([0.05+i*dw, (0.95-dh)-j*dh, dw, dh]) for i in range(ncols) if (j in [0, nrows-1] or i in [0, ncols-1])
    ] for j in range(nrows)
]

ax_corner = [ax[0][0], ax[0][ncols-1], ax[nrows-1][ncols-1], ax[nrows-1][0]]
for axi in ax_corner:
    axi.axis('off')

ax_order = [
    ax[2][1], ax[1][1],
    ax[0][3], ax[0][2], ax[0][1],
    ax[1][0], ax[2][0], ax[3][0],
    ax[4][1], ax[4][2], ax[4][3],
    ax[3][1]
]

#********************
#MAKE CENTRAL PLOT
#********************
dm = 0.0
axm = fig.add_axes([0.05+dw+dm, 0.05+dh+dm, 3*dw-2*dm, 3*dh-2*dm])

kwargs_im = dict(cmap=cmap_mom, extent=extent, levels=levels_im)
kwargs_cc = dict(colors='k', linestyles='-', extent=extent, levels=levels_cc, linewidths=0.4)
kwargs_cbar = dict(orientation='horizontal', pad=0.03, shrink=0.95, aspect=15)

im0 = axm.contourf(map2d, extend='both', **kwargs_im)
if args.show_contours:
    cc0 = axm.contour(map2d, **kwargs_cc)

make_up_ax(axm, xlims=(-xlim, xlim), ylims=(-xlim, xlim), labelsize=11)
axm.axis('off')

model.make_emission_surface(
    axm,
    R_lev=R_lev,
    #which='upper',
    kwargs_R={'colors': '0.1', 'linewidths': 0.5},
    kwargs_phi={'colors': '0.1', 'linewidths': 0.5}
)
model.make_disc_axes(axm)

#**************************
#GET AND PLOT LINE PROFILE
#**************************
get_sky_from_disc_coords = GridTools.get_sky_from_disc_coords

incl = model.params['orientation']['incl']
PA = model.params['orientation']['PA']
xc = model.params['orientation']['xc']
yc = model.params['orientation']['yc']
vsys = model.params['velocity']['vsys']
orient = (incl, PA, xc, yc)

get_z = lambda r: model.z_upper_func({'R': np.asarray(r)*au_to_m}, **model.params['height_upper'])/au_to_m

def get_lineprofile(data, r, az, npix=0, avg_func=np.nanmean):
    z = get_z(r)
    x, y, _ = get_sky_from_disc_coords(r, az, z, *orient)
    j = np.argmin(np.abs(model.skygrid['xygrid'][0]-x*au_to_m))
    i = np.argmin(np.abs(model.skygrid['xygrid'][1]-y*au_to_m))
    if npix>0:
        line = [avg_func(chan[i-npix:i+npix+1, j-npix:j+npix+1]) for chan in data]
        return x, y, np.asarray(line)
    else:
        return x, y, data[:,i,j]

kw_spec = dict(where="mid", linewidth=1.4, color='k')
cspine = '0.55'
line_peaks, vel_peaks = [], []

#**********************************
#READ PARCUBE AND MAKE CUBE FROM IT
#**********************************
if 'double' in args.kernel:
    parcube_up = fits.getdata('parcube_up_%s_%s_%s.fits'%(args.kernel, args.kind, args.type))
    parcube_low = fits.getdata('parcube_low_%s_%s_%s.fits'%(args.kernel, args.kind, args.type))
    fitdata_up = fit_kernel.get_channels_from_parcube(parcube_up, None, vchannels, method=args.kernel, kind=args.kind)
    fitdata_low = fit_kernel.get_channels_from_parcube(None, parcube_low, vchannels, method=args.kernel, kind=args.kind)        
else:
    parcube_up = fits.getdata('parcube_%s_%s.fits'%(args.kernel, args.type))
    parcube_low = None
            
fitdata = fit_kernel.get_channels_from_parcube(parcube_up, parcube_low, vchannels, method=args.kernel, kind=args.kind)
fitcube = Cube(fitdata, cube.header, vchannels, dpc, beam=cube.beam)

#***********************************
#MAKE LINE PROFILE FOR EACH SUBPANEL
#***********************************
for k, az in enumerate(np.arange(0, 2*np.pi, np.pi/6)):
    axi = ax_order[k]
    az_deg = np.degrees(az) + args.phi0
    az_rad = np.radians(az_deg)
    text = k 

    if args.showfit:
        xm, ym, linem = get_lineprofile(fitcube.data, args.radius, az_rad, npix=args.npix)    
        axi.plot(vchannels, linem, lw=3.5, color='orange') #color='limegreen') #color='mediumorchid') #color='0.6') #

    x, y, line = get_lineprofile(cube.data, args.radius, az_rad, npix=args.npix)        
    axi.step(vchannels, line, **kw_spec)

    if args.showfit and 'double' in args.kernel:
        xu, yu, lineu = get_lineprofile(fitdata_up, args.radius, az_rad, npix=args.npix)
        xl, yl, linel = get_lineprofile(fitdata_low, args.radius, az_rad, npix=args.npix)
        axi.fill_between(vchannels, linel, color='magenta', step='mid', alpha=0.3, hatch='////')
        axi.fill_between(vchannels, lineu, color='lime', step='mid', alpha=0.3)
        
    axi.text(0.9, 0.9, text, va='top', ha='right', fontsize=13, color='k', fontweight='bold', transform=axi.transAxes)
    axi.axvspan(vsys, vsys+10, facecolor='tomato', alpha=0.1)
    axi.axvspan(vsys-10, vsys, facecolor='dodgerblue', alpha=0.1)
    
    axi.axvline(vsys, color='0.2', lw=1.5, dash_capstyle='round', dashes=(0.5, 1.5), alpha=0.8)
    if args.channel_id > -1:
        axi.axvline(cube.vchannels[args.channel_id], linestyle='--', color='cyan', lw=1.5)            
    
    for spine in ['left',  'top', 'right', 'bottom']:
        axi.spines[spine].set_color(cspine)        
    axi.tick_params(which='both', color=cspine, labelcolor=cspine)
    
    if args.npix>0:
        rect = patches.Rectangle(
            (x-step*args.npix, y-step*args.npix),
            2*step*args.npix,
            2*step*args.npix,
            lw=1.5,
            edgecolor='k',
            facecolor="none",
        )
        axm.add_patch(rect)
        dxt = step*np.cos(az_rad+PA)*(4+args.npix)
        dyt = step*np.sin(az_rad+PA)*(4+args.npix)        
        axm.text(x+dxt, y+dyt, text, va='center', ha='center', color='k', fontsize=13)

    else:
        axm.text(x, y, text, va='center', ha='center', color='k', fontsize=13)

    line_peaks.append(np.nanmax(line))
    vel_peaks.append(vchannels[np.argmax(line)])
        
disclabel = plot_title.split('$-$')[0]
momentlabel = plot_title.split('$-$')[1]    
axm.text(0.001, 0.98, momentlabel, va='top', ha='left', color='0.1', fontsize=14, transform=axm.transAxes)
axm.text(0.98, 0.02, disclabel, va='bottom', ha='right', color='0.1', fontsize=14, transform=axm.transAxes)
cube.plot_beam(axm, fc='0.8')
    
max_peak = np.max(line_peaks)        
min_vel = np.min(vel_peaks)
max_vel = np.max(vel_peaks)

#************************
#DECORATE SUBPLOTS LAYOUT
#************************
kw_ax = dict(labelsize=11)
for row, axrow in enumerate(ax):
    for col, axi in enumerate(axrow):
        
        if row==0 and col==1:
            make_up_ax(axi, labelbottom=False, labeltop=True, **kw_ax)
        if row==0 and col!=1:
            make_up_ax(axi, labelbottom=False, labeltop=False, labelleft=False, **kw_ax)
            
        if row>0 and col==0:
            make_up_ax(axi, labelbottom=False, labeltop=False, labelleft=False, **kw_ax)
        if row==nrows-2 and col==0:
            make_up_ax(axi, labelbottom=True, labeltop=False, labelleft=True, **kw_ax)            
            axi.set_xlabel(r'$\upsilon_{\rm l.o.s}$ [km/s]', fontsize=13, color=cspine)
            axi.set_ylabel(r'Intensity [K]', fontsize=13, color=cspine)            

        if row==1 and col==1:
            make_up_ax(axi, labelbottom=False, labeltop=True, labelleft=False, labelright=True, **kw_ax)                        
        if row>1 and col==1:
            make_up_ax(axi, labelbottom=False, labeltop=False, labelleft=False, labelright=False, **kw_ax)
            
        if row==nrows-1 and col==ncols-2:
            make_up_ax(axi, labelbottom=True, labeltop=False, labelleft=False, labelright=True, **kw_ax)                        
        if row==nrows-1 and col<ncols-2:
            make_up_ax(axi, labelbottom=False, labeltop=False, labelleft=False, labelright=False, **kw_ax)                        

        mod_major_ticks(axi, axis='x', nbins=5)
        mod_major_ticks(axi, axis='y', nbins=4)

        axi.set_xlim(min_vel-1.5, max_vel+1.5)
        axi.set_ylim(-0.1*max_peak, 1.1*max_peak)

        axi.tick_params(labelsize=12)
        
plt.savefig('line_profiles_%s_%s_%s_%.0fau_phi0_%sdeg_%s.png'%(meta['disc'], meta['mol'],args.kernel,args.radius,args.phi0,args.type), bbox_inches='tight', dpi=200)    
show_output(args)



from discminer.core import Data
from discminer.rail import Contours
from discminer.plottools import (make_up_ax,
                                 add_cbar_ax,
                                 mod_major_ticks,
                                 mod_minor_ticks,
                                 mod_nticks_cbars,
                                 use_discminer_style)

from utils import (get_2d_plot_decorators,
                   init_data_and_model,
                   get_noise_mask,
                   load_moments,
                   load_disc_grid,
                   add_parser_args)

import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u
from astropy.io import fits

import json
from argparse import ArgumentParser

use_discminer_style()

parser = ArgumentParser(prog='Comparison between moments from different kernels', description='')
parser.add_argument('-t', '--component', default='double', type=str, choices=['single', 'double'], help="Single or double component moment")
args = add_parser_args(parser, moment=True, surface=True, kind=True, Rinner=True, Router=True, smooth=0.5)
     
#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']
vsys = best['velocity']['vsys']
Rout = best['intensity']['Rout']

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment, fmt_vertical=True)

chot = '0.5'
cmap_mom.set_under(chot)
cmap_mom.set_over(chot)

cmap_res.set_under(chot)
cmap_res.set_over(chot)

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']

dpc = meta['dpc']*u.pc
Rmax = 1.1*Rout*u.au #Max model radius, 10% larger than disc Rout

#********************
#LOAD DATA AND GRID
#********************
datacube, model = init_data_and_model(Rmin=0, Rmax=1.1)
noise_mean, mask = get_noise_mask(datacube, thres=3)

#Useful definitions for plots
with open('grid_extent.json') as json_file:
    grid = json.load(json_file)

xmax = grid['xsky'] 
xlim = np.min([xmax, Rmax.value])*1.0
extent= np.array([-xmax, xmax, -xmax, xmax])

#model.make_model()
#*************************
#LOAD DISC GEOMETRY
R, phi, z = load_disc_grid()

#**************************
#MAKE PLOT
nrows=3
if args.component=='double':    
    ncols=4
    kernels = ['quadratic', 'gaussian', 'dgauss', 'dbell']
elif args.component=='single':
    ncols=3
    kernels = ['quadratic', 'gaussian', 'bell']

figx, figy = 2.5*ncols+1, nrows*2+1
    
fig, ax = plt.subplots(ncols=ncols, nrows=nrows, figsize=(figx, figy))
plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=0.10, hspace=0.0)

kwargs_im = dict(cmap=cmap_mom, extent=extent, levels=levels_im)
kwargs_cc = dict(colors='k', linestyles='-', extent=extent, levels=levels_cc, linewidths=0.4)

for i,kernel in enumerate(kernels):
    
    try:
        moment_data, moment_model, residuals, mtags = load_moments(
            args,
            kernel=kernel,
            mask=mask,
            clip_Rmin=args.Rinner*datacube.beam_size,
            clip_Rmax=args.Router*Rout*u.au,
            clip_Rgrid=R[args.surface]*u.m
        )
    except FileNotFoundError:
        continue

    im0 = ax[0,i].contourf(moment_data, extend='both', **kwargs_im)
    im1 = ax[1,i].contourf(moment_model, extend='both', **kwargs_im)
    im2 = ax[2,i].contourf(residuals, cmap=cmap_res, origin='lower', extend='both', extent=extent, levels=np.linspace(-1.05*clim, 1.05*clim, 32))

    cc0 = ax[0,i].contour(moment_data, **kwargs_cc)
    cc1 = ax[1,i].contour(moment_model, **kwargs_cc)

    ax[0,i].set_title(kernel.capitalize(), pad=15, color='0.7', fontsize=15)

ax[0,0].set_ylabel(ctitle, fontsize=14)    
ax[1,0].set_ylabel('Model', fontsize=14)
ax[2,0].set_ylabel('Residuals', fontsize=14)

#*************
#COLORBARS
#*************
kwargs_addax = dict(perc=9, pad=0.0, orientation='vertical') 
ax_cbar0 = add_cbar_ax(fig, ax[0,-1], **kwargs_addax)
ax_cbar1 = add_cbar_ax(fig, ax[1,-1], **kwargs_addax)
ax_cbar2 = add_cbar_ax(fig, ax[2,-1], **kwargs_addax)

axp0 = ax_cbar0.get_position()
axp1 = ax_cbar1.get_position()
axp2 = ax_cbar2.get_position()

ax_cbar01 = fig.add_axes([axp1.x0, axp1.y0, axp1.width, axp0.y1-axp1.y0])
ax_cbar0.remove()
ax_cbar1.remove()

kwargs_cbar = dict(orientation='vertical')

cbar1 = plt.colorbar(im1, cax=ax_cbar01, format=cfmt, **kwargs_cbar)
cbar1.ax.tick_params(labelsize=8) 
cbar2 = plt.colorbar(im2, cax=ax_cbar2, format=cfmt, **kwargs_cbar)
cbar2.ax.tick_params(labelsize=8) 

mod_nticks_cbars([cbar1], nbins=10)
mod_nticks_cbars([cbar2], nbins=5)

cbar1.set_label(clabel, fontsize=9)
cbar2.set_label(r'%s'%unit, fontsize=9)

#****************
#AXES DECORATIONS
#****************
kw_tickpars = dict(labelleft=False, labeltop=False, left=False, right=False, bottom=False, top=False)

for i,axrow in enumerate(ax):
    for axi in axrow:
        make_up_ax(axi, xlims=(-xlim, xlim), ylims=(-xlim, xlim))
        mod_major_ticks(axi, axis='both', nbins=8)
        axi.set_aspect(1)
        axi.tick_params(**kw_tickpars)
        #if i==nrows-1: axi.set_xlabel('Offset [au]', fontsize=11)
        
        #model.make_emission_surface(
        Contours.emission_surface(
            axi,
            R, phi, extent=extent,
            R_lev=np.linspace(0.1, 0.97, 10)*Rout*u.au.to('m'),
            which='both',
            kwargs_R={'colors': '0.1', 'linewidths': 0.1, 'linestyles': '-'},
            kwargs_phi={'colors': '0.1', 'linewidths': 0.05, 'linestyles': '-'}
        )
        #model.make_disc_axes(axi)        

for axi in ax[0]:
    datacube.plot_beam(axi, fc='lime')
    
plt.savefig('moment+comparison_%s.png'%args.moment, dpi=200, bbox_inches='tight')
plt.show()
plt.close()



from discminer.mining_control import _mining_isovelocities
from discminer.mining_utils import (get_2d_plot_decorators,
                                    get_noise_mask,
                                    load_moments,
                                    load_disc_grid,
                                    overlay_continuum,
                                    overlay_filaments,
                                    overlay_spirals,
                                    mark_planet_location,
                                    show_output)

from discminer.core import Data
import discminer.cart as cart
from discminer.plottools import (make_round_map,
                                 make_polar_map,
                                 make_substructures,
                                 make_up_ax,
                                 mod_major_ticks,
                                 mod_nticks_cbars,
                                 use_discminer_style,
                                 get_cmap_from_color,
                                 get_continuous_cmap)

import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u

import copy
import json

import warnings

use_discminer_style()

if __name__ == '__main__':
    parser = _mining_isovelocities(None)
    args = parser.parse_args()

spids = np.asarray(args.spiral_ids)
spirals_pos = spids[spids>0]
spirals_neg = spids[spids<0]

#**************************
#JSON AND SOME DEFINITIONS
#**************************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

vsys = best['velocity']['vsys']
Rout = best['intensity']['Rout']
incl = best['orientation']['incl']
PA = best['orientation']['PA']
xc = best['orientation']['xc']
yc = best['orientation']['yc']

gaps = custom['gaps']
rings = custom['rings']
try:
    kinks = custom['kinks']
except KeyError:
    kinks = []
    
ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment, unit_simple=True, fmt_vertical=True)

chan_step = custom['chan_step']
nchans = custom['nchans']

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc
tag_figure = args.projection

#********************
#LOAD DATA AND GRID
#********************
datacube = Data(file_data, dpc) # Read data and convert to Cube object

#Useful definitions for plots
with open('grid_extent.json') as json_file:
    grid = json.load(json_file)

beam_au = datacube.beam_size.to('au').value
if args.absolute_Rinner>=0:
    Rmod_in = args.absolute_Rinner
else:
    Rmod_in = args.Rinner*beam_au

if args.absolute_Router>=0:
    Rmod_out = args.absolute_Router
else:
    Rmod_out = args.Router*Rout

Rmax = 1.1*Rmod_out*u.au #Max window extent, 10% larger than disc Rout
    
xmax = grid['xsky'] 
xlim = 1.15*np.min([xmax, Rmax.value])
extent= np.array([-xmax, xmax, -xmax, xmax])

#****************************
#LOAD DISC GEOMETRY AND MASK
R, phi, z = load_disc_grid()

noise_mean, mask = get_noise_mask(datacube, thres=2,
                                  mask_phi={'map2d': np.degrees(phi['upper']),
                                            'lims': args.mask_phi},
                                  mask_R={'map2d': R['upper']/au_to_m,
                                          'lims': args.mask_R}
)

Xproj = R[args.surface]*np.cos(phi[args.surface])
Yproj = R[args.surface]*np.sin(phi[args.surface])

#*************************
#LOAD AND CLIP MOMENT MAPS    
moment_data, moment_model, residuals, mtags = load_moments(
    args,
    mask=mask,
    clip_Rmin=0.0*u.au,
    clip_Rmax=Rmod_out*u.au,
    clip_Rgrid=R[args.surface]*u.m
)

#********************
#SURFACE PROPERTIES
if args.surface in ['up', 'upper']:
    if 'surf2pwl' in meta['kind']:
        z_func = cart.z_upper_powerlaw
    elif 'surfirregular' in meta['kind']:
        z_func = cart.z_upper_irregular
    else:
        z_func = cart.z_upper_exp_tapered
    z_pars = best['height_upper']
        
elif args.surface in ['low', 'lower']:
    if 'surf2pwl' in meta['kind']:
        z_func = cart.z_lower_powerlaw
    elif 'surfirregular' in meta['kind']:
        z_func = cart.z_lower_irregular            
    else:
        z_func = cart.z_lower_exp_tapered
    z_pars = best['height_lower']

#***********
#MAKE PLOT
clabels = {
    'linewidth': r'$\Delta$ Line width [km s$^{-1}$]',
    'lineslope': r'$\Delta$ Line slope',
    'velocity': r'$\Delta$ Centroid [km s$^{-1}$]',
    'peakintensity': r'$\Delta$ Peak Int. [K]'
}

idlim = int(0.5*chan_step*(nchans-1))
plot_channels = np.linspace(-idlim,idlim,nchans) + np.argmin(np.abs(datacube.vchannels-best['velocity']['vsys']))
levels = np.sort(datacube.vchannels[plot_channels.astype(int)])

if args.projection=='cartesian':
    #levels_resid = np.linspace(-clim, clim, 32)
        
    fig, ax = make_round_map(moment_data, levels, Xproj*u.m, Yproj*u.m, Rmod_out*u.au,
                             z_func=z_func, z_pars=z_pars, incl=incl, PA=PA, xc=xc, yc=yc,
                             cmap=cmap_res, clabel=unit, fmt=cfmt, 
                             gaps=gaps, rings=rings, kinks=kinks,
                             make_contourf=False, make_contour=True,
                             make_radial_grid=False, make_azimuthal_grid=True,
                             make_cbar=False,
                             mask_inner=args.Rinner*datacube.beam_size)

    X = Xproj/au_to_m
    Y = Yproj/au_to_m
    cmodel = ax.contour(X, Y, moment_model, levels=levels, colors='k', linestyles='--', linewidths=0.7)

    #SHOW FILAMENTS?
    if args.show_filaments:
        overlay_filaments(ax, Xproj/au_to_m, Yproj/au_to_m,
                          residuals, projection=args.projection, writefits=False)

    #OVERLAY FIT SPIRALS?
    if len(args.spiral_ids)>0:
        overlay_spirals(ax, args, mtags, Rmin=0, Rmax=Rmod_out)
        tag_figure += '_' + args.spiral_type + '_spirals_' + args.spiral_moment
            
    #SHOW CONTINUUM?        
    if args.show_continuum in ['all', 'scattered']:
        try:
            import cmasher as cmr
            cmap = plt.get_cmap('cmr.watermelon')
            overlay_continuum(ax, parfile='parfile_scattered.json', cmap=cmap, vmax=0.8, surface=args.surface, extend='max')
        except FileNotFoundError as e:
            warnings.warn('Unable to load parfile for scattered light image...', Warning)

    if args.show_continuum in ['all', 'band7']:
        try:
            cmap = get_cmap_from_color('orange', lev=32, vmin=0.0, vmax=0.7)
            overlay_continuum(ax, parfile='parfile_band7.json', lev=5, contours=True, cmap=cmap, surface=args.surface, zorder=20)
        except FileNotFoundError as e:
            warnings.warn('Unable to load parfile for band7 image...', Warning)

    if args.show_continuum=='none':
        make_substructures(ax, gaps=gaps, rings=rings, kinks=kinks, twodim=True, label_rings=True, label_kinks=True)
    else:
        make_substructures(ax, rings=rings, kinks=kinks, twodim=True, label_rings=True, kwargs_rings={'color': 'k', 'alpha': 0})

        
elif args.projection=='polar':
    #levels_resid = np.linspace(-clim, clim, 48)
    fig, ax, cbar = make_polar_map(moment_data, levels,
                                   R[args.surface]*u.m, phi[args.surface]*u.rad, Rmod_out*u.au,
                                   Rin=args.Rinner*datacube.beam_size,
                                   make_contourf=False, make_contour=True,                                   
                                   cmap=cmap_res, fmt=cfmt, clabel=clabels[args.moment])

    fig, ax, cbar = make_polar_map(moment_model, levels,
                                   R[args.surface]*u.m, phi[args.surface]*u.rad, Rmod_out*u.au,
                                   fig=fig, ax=ax,                                   
                                   Rin=args.Rinner*datacube.beam_size,
                                   make_contourf=False, make_contour=True,
                                   kwargs_contour={'linestyles': '--'},
                                   cmap=cmap_res, fmt=cfmt, clabel=clabels[args.moment])
    
    if args.show_filaments:
        overlay_filaments(ax, np.degrees(phi[args.surface]), R[args.surface]/au_to_m,
                          residuals, projection=args.projection, writefits=False)
        
    make_substructures(ax, gaps=gaps, rings=rings, twodim=True, polar=True, label_rings=True)

mark_planet_location(ax, args, edgecolors='limegreen', lw=4.5, s=650, alpha=0.8, coords='disc', zfunc=z_func, zpars=z_pars, incl=incl, PA=PA, xc=xc, yc=yc, dpc=dpc)        
ax.set_title(ctitle, fontsize=16, color='k')

plt.savefig('isovelocities_deproj_%s_%s.png'%(mtags['base'], tag_figure), bbox_inches='tight', dpi=200)
show_output(args)



from discminer.core import Data
from discminer.cube import Cube
from discminer.disc2d import General2d
from discminer.rail import Contours
from discminer.plottools import use_discminer_style
from discminer.tools import fit_kernel

import numpy as np
import matplotlib.pyplot as plt

from astropy import units as u
from astropy.io import fits

import json
import sys

from argparse import ArgumentParser
from utils import add_parser_args

use_discminer_style()

parser = ArgumentParser(prog='Plot reconstructed cubes', description='Plot upper+lower reconstructed cubes')
args = add_parser_args(parser, kernel='doublebell', kind=True)

#**********
#READ JSON
#**********
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc

datacube = Data(file_data, dpc)
datacube.convert_to_tb(writefits=False)

vchannels = datacube.vchannels

#Useful definitions for plots
with open('grid_extent.json') as json_file:
    grid = json.load(json_file)

xmax = grid['xsky'] 
extent= np.array([-xmax, xmax, -xmax, xmax])

#****************
#READ PARCUBE AND MAKE CUBE FROM IT
#****************
parcube_up = fits.getdata('parcube_up_%s_%s_data.fits'%(args.kernel, args.kind))
parcube_low = fits.getdata('parcube_low_%s_%s_data.fits'%(args.kernel, args.kind))

fitdata = fit_kernel.get_channels_from_parcube(parcube_up, parcube_low, vchannels, method=args.kernel, kind=args.kind, n_fit=None)
fitcube = Cube(fitdata, datacube.header, vchannels, dpc, beam=datacube.beam)

#SHOW 
fitcube.show(compare_cubes=[datacube], extent=extent, int_unit='Intensity [K]', vmin=0.0, vmax=40, show_beam=True)
fitcube.show_side_by_side(datacube, extent=extent, int_unit='Intensity [K]', vmin=0.0, vmax=40, show_beam=True)



from .utils import FrontendUtils, InputError, FITSUtils, get_tb

import numpy as np
from astropy.io import fits
from scipy.optimize import curve_fit

from tqdm import tqdm
from pathos.multiprocessing import ProcessingPool as Pool

import copy

_progress_bar = FrontendUtils._progress_bar

def _gauss(x, *p):
    A1, mu1, sigma1, *rest = p
    A2 = rest[0] if rest else 0.0
    return A1*np.exp(-(x-mu1)**2/(2.*sigma1**2)) + A2

def _doublegauss_sum(x, *p):
    A1, mu1, sigma1, A2, mu2, sigma2 = p
    return A1*np.exp(-(x-mu1)**2/(2.*sigma1**2)) + A2*np.exp(-(x-mu2)**2/(2.*sigma2**2))

def _doublegauss_mask(x, *p):
    A1, mu1, sigma1, A2, mu2, sigma2 = p
    gauss1 = A1*np.exp(-(x-mu1)**2/(2.*sigma1**2))
    gauss2 = A2*np.exp(-(x-mu2)**2/(2.*sigma2**2))
    return np.where(gauss1>=gauss2, gauss1, gauss2)

def _bell(x, *p):
    A1, mu1, sigma1, Ls1, *rest = p
    A2 = rest[0] if rest else 0.0
    return A1/(1+np.abs((x-mu1)/sigma1)**(2*Ls1)) + A2

def _doublebell_sum(x, *p):
    A1, mu1, sigma1, Ls1, A2, mu2, sigma2, Ls2 = p
    return A1/(1+np.abs((x-mu1)/sigma1)**(2*Ls1)) + A2/(1+np.abs((x-mu2)/sigma2)**(2*Ls2))

def _doublebell_mask(x, *p):
    A1, mu1, sigma1, Ls1, A2, mu2, sigma2, Ls2 = p
    bell1 = A1/(1+np.abs((x-mu1)/sigma1)**(2*Ls1))
    bell2 = A2/(1+np.abs((x-mu2)/sigma2)**(2*Ls2))
    return np.where(bell1>=bell2, bell1, bell2)

def _not_available(method):
    raise InputError(method, "requested moment method is currently unavailable")

def get_channels_from_parcube(parcube_up, parcube_low, vchannels, method='doublebell', kind='mask', n_fit=None):
    
    if parcube_up is None and 'double' in method:
        parcube_up = np.zeros_like(parcube_low)
    
    if parcube_low is None and 'double' in method:
        parcube_low = np.zeros_like(parcube_up)

    nx, ny = parcube_up.shape[1:]
    intensity = np.zeros((len(vchannels), ny, nx))
        
    par_double = lambda i,j: np.append(parcube_up[:,i,j], parcube_low[:,i,j])
    par_single = lambda i,j: parcube_up[:,i,j]
    
    if 'double' in method:
        pars = par_double
    else:
        pars = par_single

    #******************************
    #KERNELS AND RELEVANT FUNCTIONS
    if method=='doublegaussian':
        fit_func1d = _gauss
        if kind=='sum':
            fit_func = _doublegauss_sum
        elif kind=='mask':
            fit_func = _doublegauss_mask
        else:
            raise InputError(kind, "kind must be 'mask' or 'sum'")

    elif method=='doublebell':
        fit_func1d = _bell
        if kind=='sum':
            fit_func = _doublebell_sum
        elif kind=='mask':
            fit_func = _doublebell_mask
        else:
            raise InputError(kind, "kind must be 'mask' or 'sum'")

    elif method=='gaussian':
        fit_func = _gauss

    elif method=='bell':
        fit_func = _bell
        
    else:
        _not_available(method)

    #***********************
    #EVALUATE PARS ON KERNEL
    for i in range(ny):
        for j in range(nx):
            intensity[:,i,j] = fit_func(vchannels, *pars(i,j))
    
    return intensity


def fit_twocomponent(cube, model=None, lw_chans=1.0, lower2upper=1.0,
                     method='doublegaussian', kind='mask', sigma_thres=5,
                     sigma_fit=None, mask=None, planck=False,
                     niter=4, neighs=5, av_func=np.nanmedian,
                     mask_radially=True
):

    data = cube.data    
    vchannels = cube.vchannels
    dv = np.mean(vchannels[1:]-vchannels[:-1])
    lw_sign = np.sign(dv)
    
    nchan, nx, ny = np.shape(data)
    n_one, n_two, n_bad, n_hot, n_mask = 0, 0, 0, 0, 0
    n_fit = np.zeros((nx, ny))
    
    peak_up, dpeak_up = np.zeros((2, nx, ny))
    centroid_up, dcentroid_up = np.zeros((2, nx, ny))
    linewidth_up, dlinewidth_up = np.zeros((2, nx, ny))
    lineslope_up, dlineslope_up = np.zeros((2, nx, ny))    
    peak_low, dpeak_low = np.zeros((2, nx, ny))
    centroid_low, dcentroid_low = np.zeros((2, nx, ny))
    linewidth_low, dlinewidth_low = np.zeros((2, nx, ny))
    lineslope_low, dlineslope_low = np.zeros((2, nx, ny))
    
    is_dbell = method=='doublebell'
    
    #MODEL AS INITIAL GUESS?
    if model is None:
        print ('Guessing upper surface properties from data to use them as seeds for both upper (primary) and lower (secondary) surface components ...')        
        ind_max = np.nanargmax(data, axis=0)
        I_max = np.nanmax(data, axis=0)
        vel_peak = vchannels[ind_max]
        I_max_upper = I_max
        I_max_lower = lower2upper*I_max
        vel_peak_upper = vel_peak_lower = vel_peak
        lw_upper = lw_lower = lw_chans*dv*np.ones_like(vel_peak)
        ls_upper = ls_lower = 1.5
        
    else:
        vel2d, int2d, linew2d, lineb2d = model.props
        I_upper = int2d['upper']*cube.beam_area
        I_lower = int2d['lower']*cube.beam_area
        print ('Using upper and lower surface properties from discminer model as initial guesses...')
        
        if np.any(np.array(['K', 'Kelvin', 'K ', 'Kelvin ']) == cube.header['BUNIT']): #If input unit is K the raw model intensity must be converted    
            r"""
            mheader = copy.copy(cube.header)
            mheader["BUNIT"] = "beam-1 Jy"
            I_upper = FITSUtils._convert_to_tb(I_upper*cube.beam_area, mheader, cube.beam, planck=False, writefits=False)
            #"""
            restfreq = cube.header['RESTFRQ']*1e-9
            I_upper = get_tb(1e3*I_upper, restfreq, cube.beam, full=planck) #mJy/beam-->K
            I_lower = get_tb(1e3*I_lower, restfreq, cube.beam, full=planck) 

        #Initial guesses
        ind_max = np.nanargmax(data, axis=0)
        cube_max = np.take_along_axis(data, ind_max[None], axis=0).squeeze()
        I_max_upper = np.where(np.isnan(I_upper), 1.0*cube_max, I_upper)
        I_max_lower = np.where(np.isnan(I_lower), 0.5*cube_max, I_lower)
        vel_peak_upper = np.where(np.isnan(vel2d['upper']), vchannels[ind_max], vel2d['upper'])
        vel_peak_lower = np.where(np.isnan(vel2d['lower']), vchannels[ind_max], vel2d['lower'])
        lw_upper = lw_sign*np.where(np.isnan(linew2d['upper']), lw_sign*1.5*dv, linew2d['upper'])
        lw_lower = lw_sign*np.where(np.isnan(linew2d['lower']), lw_sign*1.5*dv, linew2d['lower'])
        ls_upper = np.where(np.isnan(linew2d['upper']), 1.5, 1*lineb2d['upper']) #0.5*
        ls_lower = np.where(np.isnan(linew2d['lower']), 1.5, 1*lineb2d['lower']) #0.5*

    if sigma_fit is None: sigma_func = lambda i,j: None
    else: sigma_func = lambda i,j: sigma_fit[:,i,j]

    noise = np.std( np.append(data[:5,:,:], data[-5:,:,:], axis=0), axis=0) #rms intensity from first and last 5 channels

    if mask is None:
        mask = np.zeros((nx,ny)).astype(bool)

    mask = mask | (np.nanmax(data, axis=0) <= sigma_thres*noise)

    if model is not None and mask_radially:
        R, phi, z = [model.projected_coords[key] for key in ['R', 'phi', 'z']]
        mask = mask | np.isnan(R['upper'])
        
    #******************************
    #KERNELS AND RELEVANT FUNCTIONS
    if method=='doublegaussian':
        n_pars = 6
        fit_func1d = _gauss
        pfunc_two = lambda i,j: [I_max_upper[i,j], vel_peak_upper[i,j], lw_upper[i,j],
                                 I_max_lower[i,j], vel_peak_lower[i,j], lw_lower[i,j]]
        pfunc_one = lambda i,j: [I_max_upper[i,j], vel_peak_upper[i,j], lw_upper[i,j]]
        if kind=='sum':
            fit_func = _doublegauss_sum
        elif kind=='mask':
            fit_func = _doublegauss_mask
        else:
            raise InputError(kind, "kind must be 'mask' or 'sum'")
        idlow = np.array([3,4,5])
        bound0 = (0, -np.inf, -np.inf, 0, -np.inf, -np.inf)
        bound1 = (np.inf, np.inf, np.inf, np.inf, np.inf, np.inf)
        bounds = [[0, vchannels[0], 0, 0, vchannels[0], 0], [np.inf, vchannels[-1], np.abs(vchannels[-1]-vchannels[0]), np.inf, vchannels[-1], np.abs(vchannels[-1]-vchannels[0])]]
        
    elif method=='doublebell':
        n_pars = 8
        fit_func1d = _bell
        pfunc_two = lambda i,j: [I_max_upper[i,j], vel_peak_upper[i,j], lw_upper[i,j], ls_upper[i,j],
                                 I_max_lower[i,j], vel_peak_lower[i,j], lw_lower[i,j], ls_lower[i,j]]
        pfunc_one = lambda i,j: [I_max_upper[i,j], vel_peak_upper[i,j], lw_upper[i,j], ls_upper[i,j]]
        
        if kind=='sum':
            fit_func = _doublebell_sum
        elif kind=='mask':
            fit_func = _doublebell_mask
        else:
            raise InputError(kind, "kind must be 'mask' or 'sum'")
        idlow = np.array([4,5,6,7])
        bound0 = (0, -np.inf, -np.inf, 0, 0, -np.inf, -np.inf, 0)
        bound1 = (np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf)
        bounds = [[0, vchannels[0], 0, 0, 0, vchannels[0], 0, 0], [np.inf, vchannels[-1], np.abs(vchannels[-1]-vchannels[0]), 10, np.inf, vchannels[-1], np.abs(vchannels[-1]-vchannels[0]), 10]]

    else:
        _not_available(method)

    def fill_props_ij(i, j, coeff, deltas):
        peak_up[i,j] = coeff[0]
        centroid_up[i,j] = coeff[1]
        linewidth_up[i,j] = coeff[2]
        
        peak_low[i,j] = coeff[idlow[0]]
        centroid_low[i,j] = coeff[idlow[1]]
        linewidth_low[i,j] = coeff[idlow[2]]

        dpeak_up[i,j] = deltas[0]
        dcentroid_up[i,j] = deltas[1]
        dlinewidth_up[i,j] = deltas[2]

        dpeak_low[i,j] = deltas[idlow[0]]
        dcentroid_low[i,j] = deltas[idlow[1]]
        dlinewidth_low[i,j] = deltas[idlow[2]]

        if is_dbell:
            lineslope_up[i,j] = coeff[3]
            lineslope_low[i,j] = coeff[idlow[3]]
            dlineslope_up[i,j] = deltas[3]
            dlineslope_low[i,j] = deltas[idlow[3]]
            
    def fill_props(coeff, deltas):
        peak_up[:] = coeff[:,:,0]
        centroid_up[:] = coeff[:,:,1]
        linewidth_up[:] = coeff[:,:,2]
        
        peak_low[:] = coeff[:,:,idlow[0]]
        centroid_low[:] = coeff[:,:,idlow[1]]
        linewidth_low[:] = coeff[:,:,idlow[2]]

        dpeak_up[:] = deltas[:,:,0]
        dcentroid_up[:] = deltas[:,:,1]
        dlinewidth_up[:] = deltas[:,:,2]

        dpeak_low[:] = deltas[:,:,idlow[0]]
        dcentroid_low[:] = deltas[:,:,idlow[1]]
        dlinewidth_low[:] = deltas[:,:,idlow[2]]

        if is_dbell:
            lineslope_up[:] = coeff[:,:,3]
            lineslope_low[:] = coeff[:,:,idlow[3]]
            dlineslope_up[:] = deltas[:,:,3]
            dlineslope_low[:] = deltas[:,:,idlow[3]]
            
    print ('Fitting two-component function along velocity axis of the input cube...')

    #********
    #MAKE FIT    
    #********
    def process_xi(i, datai, maski, n_pars=8):

        empty = np.zeros(n_pars)
        n_one, n_two, n_bad, n_hot, n_mask = 0, 0, 0, 0, 0
        coeff, deltas, n_fit = [], [], []

        for j in range(ny): #must change by actual list of j's
            tmp_data = datai[:,j]
            peak = 2*np.nanmax(tmp_data)
            bounds[1][0] = peak
            bounds[1][idlow[0]] = peak             
            
            if maski[j]:
                n_mask += 1
                n_fit.append(-10)
                coeff.append(empty)
                deltas.append(empty)
                continue
            
            try:
                p0 = pfunc_two(i,j) #change by neighbour_guess call

                #if n_pars==8:
                #    bounds[1][3] = p0[3]*3
                #    bounds[1][idlow[3]] = p0[idlow[3]]*3
                    
                coeffj, var_matrix = curve_fit(fit_func,
                                               vchannels, tmp_data,
                                               p0=p0,
                                               #bounds=bounds, method='trf')
                                               ftol=1e-10, xtol=1e-10, gtol=1e-10, method='lm')

                deltasj = np.sqrt(np.abs(np.diag(var_matrix)))
                n_two += 1
                n_fit.append(2)
                
            except RuntimeError:
                try: 
                    coeffj, var_matrix = curve_fit(fit_func1d,
                                                   vchannels, tmp_data,
                                                   p0=pfunc_one(i,j)
                    )
                    coeffj = np.append(coeffj, coeffj)
                    deltasj = np.sqrt(np.abs(np.diag(var_matrix)))
                    deltasj = np.append(deltasj, deltasj)
                    n_one += 1
                    n_fit.append(1)
                    
                except RuntimeError:
                    n_bad += 1
                    n_fit.append(0)
                    coeff.append(empty)
                    deltas.append(empty)
                    continue

            coeff.append(coeffj) 
            deltas.append(deltasj)
            
            #fill_props(i,j,coeff,deltas)
        return coeff, deltas, n_fit, [n_one, n_two, n_bad, n_hot, n_mask]

    
    def parallel_processing(data, mask, n_pars=8):

        with Pool(ncpus=None) as pool:
            results = list(tqdm(pool.imap(
                lambda i: process_xi(i, data[:,i,:], mask[i,:], n_pars=n_pars),
                range(nx)
            ), total=nx, desc="Processing two-component fit"))

        coeff, deltas, n_fit, n_diag = [], [], [], []
        #phii, residi = [], []
        for coeff_local, deltas_local, n_fit_local, n_diag_local in results:
            coeff.append(coeff_local)
            deltas.append(deltas_local)
            n_fit.append(n_fit_local)
            n_diag.append(n_diag_local)

        coeff = np.asarray(coeff)
        deltas = np.asarray(deltas)
        n_fit = np.asarray(n_fit)
        n_diag = np.asarray(n_diag)

        return coeff, deltas, n_fit, n_diag
    
    coeff, deltas, n_fit, n_diag = parallel_processing(data, mask, n_pars=n_pars)
    n_one, n_two, n_bad, n_hot, n_mask = np.sum(n_diag, axis=0)

    fill_props(coeff, deltas) #If ny is not uniform, as is the case for hot pixel fit, call fill_props_ij instead, so that one can loop over ij
    
    #*****************************
    #KEEP TRACK OF 'HOT' PIXELS
    #*****************************
    #Hot pixels will be tagged as -1

    def clean_nfit():
        mm = n_fit == -10 #noise
        ii = ((peak_up < 0.0) | (peak_low < 0.0)) & (~mm) #negative intensities
        peak_thres = 2*np.nanmax(cube_max)
        jj = ((peak_up > peak_thres) | (peak_low > peak_thres)) & (~mm) #too large intensities
        cc = (centroid_up == centroid_low) & (n_fit != 1) & (~mm) #up==low
        ww = ((np.abs(linewidth_up) <= 0.5*np.abs(dv)) | (np.abs(linewidth_low) <= 0.5*np.abs(dv))) & (~mm) #component too narrow (<=half a chan width)
        dd = ((np.abs(linewidth_up) > 5.0) | (np.abs(linewidth_low) > 5.0)) & (~mm) #Unrealistically broad component

        bb = False #dcentroid_up > 1*np.abs(dv)

        w3 = False #((np.abs(linewidth_up) <= 1*np.abs(dv)) | (np.abs(linewidth_low) <= 1*np.abs(dv))) & (~mm) #narrow component
        w4 = False #(np.abs(linewidth_low) >= 2*np.abs(linewidth_up)) & (~mm) #lower line width much larger than upper

        w5 = False #dcentroid_up > 1
        
        n_fit[ii+jj+cc+ww+dd+w5] = -1 #bad pixel
        n_hot = np.sum(ii+jj+cc+ww+dd + w3+w4+w5)

        n_fit[w3] = 3 #1chan narrow component
        n_fit[w4] = 4 #1chan narrow component
        n_fit[w5] = 5 #Bad delta
        
        return n_hot
        
    #**********
    #PACK PROPS
    #**********
    upper = [peak_up, centroid_up, linewidth_up]
    dupper = [dpeak_up, dcentroid_up, dlinewidth_up]
    lower = [peak_low, centroid_low, linewidth_low]
    dlower = [dpeak_low, dcentroid_low, dlinewidth_low]

    if is_dbell:
        upper += [lineslope_up]
        dupper += [dlineslope_up]
        lower += [lineslope_low]
        dlower += [dlineslope_low]

    n_hot = clean_nfit()
    
    print ('\nTwo-component fit did not converge for %.2f%s of the pixels'%(100.0*(n_bad)/(nx*ny),'%'))
    print ('A single component was fit for %.2f%s of the pixels'%(100.0*(n_one)/(nx*ny),'%'))
    print ('Masked pixels below intensity threshold: %.2f%s'%(100.0*(n_mask)/(nx*ny),'%'))
    print ('Hot pixels: %.2f%s'%(100.0*(n_hot)/(nx*ny),'%'))        

    #******************
    #HANDLE HOT PIXELS
    #******************    
    if niter>0:
        print ('\nRe-doing fit for  %d hot pixels and %d single-component pixels'%(n_hot, n_one))

        def neighbour_guess(i, j, n_fit, neighs=3, av_func=np.nanmean):

            if i<neighs or j<neighs:
                return None
            
            neigh_arr = np.arange(-neighs, neighs+1)

            ileft, iright = i-neighs, i+neighs+1
            jleft, jright = j-neighs, j+neighs+1
            window = n_fit[ileft:iright, jleft:jright]

            masked = window==-10
            n_mask = np.sum(masked) #n masked pixels

            hot = window==-1
            n_hot = np.sum(hot) #n hot pixels
            
            one = window==1
            n_one = np.sum(one) #single-component pixels

            
            three = window==3
            n_three = np.sum(three) #narrow components

            four = window==4
            n_four = np.sum(four) #broad lower linewidth
            
            
            n_bad = n_mask+n_hot+n_one + n_three+n_four
            tot = (2*neighs+1)**2

            if tot-n_bad < n_bad:
                return None

            ic, jc = (~masked & ~hot & ~one  &  ~three & ~four).nonzero() #get clean pixels where double fit worked

            up_guess = [av_func(up_prop[ileft:iright, jleft:jright][ic, jc]) for up_prop in upper]
            low_guess = [av_func(low_prop[ileft:iright, jleft:jright][ic, jc]) for low_prop in lower]
            return up_guess+low_guess
                         
        for ni in range(niter):
            print ('Iteration #%d...'%(ni+1))

            #Select hot pixels (-1 flag) and single-component pixels (+1 flag)
            m, n = np.where((n_fit==-1) | (n_fit==1) | (n_fit>2))
            #m, n = np.where(n_fit==-1)
                
            for k in range(len(m)):
                i, j = m[k], n[k]                                
                uplow = neighbour_guess(i, j, n_fit, neighs=neighs, av_func=av_func)
                
                if uplow is None:
                    continue

                tmp_data = data[:,i,j]

                try:                    
                    coeff, var_matrix = curve_fit(fit_func,
                                                  vchannels, tmp_data,
                                                  p0 = uplow,
                                                  #bounds=bounds, method='trf')                                                  
                                                  ftol=1e-10, xtol=1e-10, gtol=1e-10, method='lm')
                    deltas = np.sqrt(np.abs(np.diag(var_matrix)))
                    n_two += 1
                    n_fit[i,j] = 2
                    
                    r"""
                    if (deltas[:3]==np.inf).any(): # or deltas[1] > 0.5*np.abs(dv):
                        coeff, var_matrix = curve_fit(fit_func1d,
                                                      vchannels, tmp_data,
                                                      p0=pfunc_one(i,j)
                        )
                        coeff = np.append(coeff, coeff)
                        deltas = np.sqrt(np.abs(np.diag(var_matrix)))
                        deltas = np.append(deltas, deltas)
                        n_one += 1
                        n_fit[i,j] = 1
                    else:                
                        n_two += 1
                        n_fit[i,j] = 2
                    #"""
                except RuntimeError:
                    try: 
                        coeff, var_matrix = curve_fit(fit_func1d,
                                                      vchannels, tmp_data,
                                                      p0=pfunc_one(i,j)
                        )
                        coeff = np.append(coeff, coeff)
                        deltas = np.sqrt(np.abs(np.diag(var_matrix)))
                        deltas = np.append(deltas, deltas)
                        n_one += 1
                        n_fit[i,j] = 1
                    
                    except RuntimeError:
                        n_bad += 1
                        n_fit[i,j] = 0                    
                        continue

                fill_props_ij(i,j,coeff,deltas)
                
            n_hot = clean_nfit()
            print ('Resulting hot pixels:', n_hot)

        #***************************************************
        #FIT ONE COMPONENT FUNCTION TO REMAINING HOT PIXELS
        #***************************************************
        m, n = np.where(n_fit==-1)
        #r"""
        for k in range(len(m)):
            i, j = m[k], n[k]                                
            tmp_data = data[:,i,j]

            try: 
                coeff, var_matrix = curve_fit(fit_func1d,
                                              vchannels, tmp_data,
                                              p0=pfunc_one(i,j)
                )
                coeff = np.append(coeff, coeff)
                deltas = np.sqrt(np.abs(np.diag(var_matrix)))
                deltas = np.append(deltas, deltas)
                n_one += 1
                n_fit[i,j] = 1
                    
            except RuntimeError:
                n_bad += 1
                n_fit[i,j] = 0                    
                continue

            fill_props_ij(i,j,coeff,deltas)
        #"""
    n_hot = clean_nfit()
    print ('Resulting hot pixels after fitting 1D component:', n_hot)
    
    return upper, dupper, lower, dlower, n_fit
    
    """
    #****************
    #TREAT HOT PIXELS
    #****************    
    def neighbour_guess(i, j, n_fit, neighs=3, av_func=np.nanmean):

        if i<neighs or j<neighs:
            return None
        
        neigh_arr = np.arange(-neighs, neighs+1)

        ileft, iright = i-neighs, i+neighs+1
        jleft, jright = j-neighs, j+neighs+1
        window = n_fit[ileft:iright, jleft:jright]
        
        masked = window==-10
        n_mask = np.sum(masked) #n masked pixels

        hot = window==-1
        n_hot = np.sum(hot) #n hot pixels
            
        one = window==1
        n_one = np.sum(one) #single-component pixels
            
        three = window==3
        n_three = np.sum(three) #narrow components

        four = window==4
        n_four = np.sum(four) #broad lower linewidth
        
        n_bad = n_mask+n_hot+n_one + n_three+n_four
        tot = (2*neighs+1)**2

        if tot-n_bad < n_bad:
            return None

        ic, jc = (~masked & ~hot & ~one  &  ~three & ~four).nonzero() #get clean pixels where double fit worked

        up_guess = [av_func(up_prop[ileft:iright, jleft:jright][ic, jc]) for up_prop in upper]
        low_guess = [av_func(low_prop[ileft:iright, jleft:jright][ic, jc]) for low_prop in lower]
        return up_guess+low_guess

    #Process rows in parallel    
    def process_ij(i, datai, bad_j_indices, n_fit, n_pars=8):
        empty = np.zeros(n_pars)
        coeff, deltas, n_fit_tmp = [], [], []
    
        for j in bad_j_indices:

            uplow = neighbour_guess(i, j, n_fit, neighs=neighs, av_func=av_func)
            
            if uplow is None:
                n_fit_tmp.append(0)
                coeff.append(empty)
                deltas.append(empty)
                continue

            tmp_data = datai[:,j]
            
            try:                    
                coeffj, var_matrix = curve_fit(fit_func,
                                               vchannels, tmp_data,
                                               p0 = uplow,
                                               ftol=1e-10, xtol=1e-10, gtol=1e-10, method='lm')
                deltasj = np.sqrt(np.abs(np.diag(var_matrix)))
                n_fit_tmp.append(2)
                    
            except RuntimeError:
                try: 
                    coeffj, var_matrix = curve_fit(fit_func1d,
                                                  vchannels, tmp_data,
                                                  p0=pfunc_one(i,j)
                    )
                    coeffj = np.append(coeffj, coeffj)
                    deltasj = np.sqrt(np.abs(np.diag(var_matrix)))
                    deltasj = np.append(deltasj, deltasj)
                    n_fit_tmp.append(1)
                    
                except RuntimeError:
                    n_fit_tmp.append(0)
                    coeff.append(empty)
                    deltas.append(empty)
                    continue            

            coeff.append(coeffj) 
            deltas.append(deltasj)
                
        return coeff, deltas, n_fit_tmp

    def parallel_processing_ij(data, bad_pixel_indices, n_fit, n_pars=8):
        rows_with_bad_fits = {}
        for i, j in bad_pixel_indices:
            if i not in rows_with_bad_fits:
                rows_with_bad_fits[i] = []
            rows_with_bad_fits[i].append(j)

        rows2fit = list(rows_with_bad_fits.keys())
        
        with Pool(ncpus=None) as pool:
            results = list(tqdm(pool.imap(
                lambda i: process_ij(i, data[:, i, :], rows_with_bad_fits[i], n_fit, n_pars=n_pars),
                rows2fit
            ), total=len(rows2fit), desc="Re-fitting hot pixels"))
            
        coeff, deltas, n_fit_tmp = [], [], []
    
        # Collect results from each row
        for coeff_local, deltas_local, n_fit_local in results:
            coeff.append(coeff_local)
            deltas.append(deltas_local)
            n_fit_tmp.append(n_fit_local)

        return coeff, deltas, n_fit_tmp
    
    if niter>0:
        print ('\nRe-doing fit for  %d hot pixels and %d single-component pixels'%(n_hot, n_one))
        
        for n in range(niter):
            print ('Iteration #%d...'%(n+1))

            #Select hot pixels (-1 flag) and single-component pixels (+1 flag)
            ib,jb = np.where((n_fit==-1) | (n_fit==1) | (n_fit>2))
            bad_pixel_indices = zip(ib,jb)
            
            coeffk, deltask, n_fitk = parallel_processing_ij(data, bad_pixel_indices, n_fit)
            
            #Update coeff, deltas, and n_fit matrices per ij
            k=0
            for i in range(len(coeffk)):
                for j in range(len(coeffk[i])):
                    fill_props_ij(ib[k], jb[k], coeffk[i][j], deltask[i][j])
                    n_fit[ib[k],jb[k]] = n_fitk[i][j]
                    k+=1
                    
            n_hot = clean_nfit()
            print ('Resulting hot pixels:', n_hot)
    
    return upper, dupper, lower, dlower, n_fit
    """

def fit_twocomponent_(cube, model=None, lw_chans=1.0, lower2upper=1.0,
                     method='doublegaussian', kind='mask', sigma_thres=5,
                     sigma_fit=None, mask=None, planck=False,
                     niter=4, neighs=5, av_func=np.nanmedian,
                     mask_radially=True
):

    data = cube.data    
    vchannels = cube.vchannels
    dv = np.mean(vchannels[1:]-vchannels[:-1])
    lw_sign = np.sign(dv)
    
    nchan, nx, ny = np.shape(data)
    n_one, n_two, n_bad, n_hot, n_mask = 0, 0, 0, 0, 0
    n_fit = np.zeros((nx, ny))
    
    peak_up, dpeak_up = np.zeros((2, nx, ny))
    centroid_up, dcentroid_up = np.zeros((2, nx, ny))
    linewidth_up, dlinewidth_up = np.zeros((2, nx, ny))
    lineslope_up, dlineslope_up = np.zeros((2, nx, ny))    
    peak_low, dpeak_low = np.zeros((2, nx, ny))
    centroid_low, dcentroid_low = np.zeros((2, nx, ny))
    linewidth_low, dlinewidth_low = np.zeros((2, nx, ny))
    lineslope_low, dlineslope_low = np.zeros((2, nx, ny))
    
    is_dbell = method=='doublebell'
    
    #MODEL AS INITIAL GUESS?
    if model is None:
        print ('Guessing upper surface properties from data to use them as seeds for both upper (primary) and lower (secondary) surface components ...')        
        ind_max = np.nanargmax(data, axis=0)
        I_max = np.nanmax(data, axis=0)
        vel_peak = vchannels[ind_max]
        I_max_upper = I_max
        I_max_lower = lower2upper*I_max
        vel_peak_upper = vel_peak_lower = vel_peak
        lw_upper = lw_lower = lw_chans*dv*np.ones_like(vel_peak)
        ls_upper = ls_lower = 1.5
        
    else:
        vel2d, int2d, linew2d, lineb2d = model.props
        I_upper = int2d['upper']*cube.beam_area
        I_lower = int2d['lower']*cube.beam_area
        print ('Using upper and lower surface properties from discminer model as initial guesses...')
        
        if np.any(np.array(['K', 'Kelvin', 'K ', 'Kelvin ']) == cube.header['BUNIT']): #If input unit is K the raw model intensity must be converted    
            r"""
            mheader = copy.copy(cube.header)
            mheader["BUNIT"] = "beam-1 Jy"
            I_upper = FITSUtils._convert_to_tb(I_upper*cube.beam_area, mheader, cube.beam, planck=False, writefits=False)
            #"""
            restfreq = cube.header['RESTFRQ']*1e-9
            I_upper = get_tb(1e3*I_upper, restfreq, cube.beam, full=planck) #mJy/beam-->K
            I_lower = get_tb(1e3*I_lower, restfreq, cube.beam, full=planck) 

        #Initial guesses
        ind_max = np.nanargmax(data, axis=0)
        cube_max = np.take_along_axis(data, ind_max[None], axis=0).squeeze()
        I_max_upper = np.where(np.isnan(I_upper), 1.0*cube_max, I_upper)
        I_max_lower = np.where(np.isnan(I_lower), 0.5*cube_max, I_lower)
        vel_peak_upper = np.where(np.isnan(vel2d['upper']), vchannels[ind_max], vel2d['upper'])
        vel_peak_lower = np.where(np.isnan(vel2d['lower']), vchannels[ind_max], vel2d['lower'])
        lw_upper = lw_sign*np.where(np.isnan(linew2d['upper']), lw_sign*1.5*dv, linew2d['upper'])
        lw_lower = lw_sign*np.where(np.isnan(linew2d['lower']), lw_sign*1.5*dv, linew2d['lower'])
        ls_upper = np.where(np.isnan(linew2d['upper']), 1.5, 1*lineb2d['upper']) #0.5*
        ls_lower = np.where(np.isnan(linew2d['lower']), 1.5, 1*lineb2d['lower']) #0.5*

    if sigma_fit is None: sigma_func = lambda i,j: None
    else: sigma_func = lambda i,j: sigma_fit[:,i,j]

    noise = np.std( np.append(data[:5,:,:], data[-5:,:,:], axis=0), axis=0) #rms intensity from first and last 5 channels

    if mask is None:
        mask = np.zeros((nx,ny)).astype(bool)

    mask = mask | (np.nanmax(data, axis=0) <= sigma_thres*noise)

    if model is not None and mask_radially:
        R, phi, z = [model.projected_coords[key] for key in ['R', 'phi', 'z']]
        mask = mask | np.isnan(R['upper'])
        
    #******************************
    #KERNELS AND RELEVANT FUNCTIONS
    if method=='doublegaussian':
        fit_func1d = _gauss
        pfunc_two = lambda i,j: [I_max_upper[i,j], vel_peak_upper[i,j], lw_upper[i,j],
                                 I_max_lower[i,j], vel_peak_lower[i,j], lw_lower[i,j]]
        pfunc_one = lambda i,j: [I_max_upper[i,j], vel_peak_upper[i,j], lw_upper[i,j]]
        if kind=='sum':
            fit_func = _doublegauss_sum
        elif kind=='mask':
            fit_func = _doublegauss_mask
        else:
            raise InputError(kind, "kind must be 'mask' or 'sum'")
        idlow = np.array([3,4,5])
        bound0 = (0, -np.inf, -np.inf, 0, -np.inf, -np.inf)
        bound1 = (np.inf, np.inf, np.inf, np.inf, np.inf, np.inf)

    elif method=='doublebell':
        fit_func1d = _bell
        pfunc_two = lambda i,j: [I_max_upper[i,j], vel_peak_upper[i,j], lw_upper[i,j], ls_upper[i,j],
                                 I_max_lower[i,j], vel_peak_lower[i,j], lw_lower[i,j], ls_lower[i,j]]
        pfunc_one = lambda i,j: [I_max_upper[i,j], vel_peak_upper[i,j], lw_upper[i,j], ls_upper[i,j]]
        
        if kind=='sum':
            fit_func = _doublebell_sum
        elif kind=='mask':
            fit_func = _doublebell_mask
        else:
            raise InputError(kind, "kind must be 'mask' or 'sum'")
        idlow = np.array([4,5,6,7])
        bound0 = (0, -np.inf, -np.inf, 0, 0, -np.inf, -np.inf, 0)
        bound1 = (np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf)

    else:
        _not_available(method)

    def fill_props(i, j, coeff):
        peak_up[i,j] = coeff[0]
        centroid_up[i,j] = coeff[1]
        linewidth_up[i,j] = coeff[2]
        
        peak_low[i,j] = coeff[idlow[0]]
        centroid_low[i,j] = coeff[idlow[1]]
        linewidth_low[i,j] = coeff[idlow[2]]

        dpeak_up[i,j] = deltas[0]
        dcentroid_up[i,j] = deltas[1]
        dlinewidth_up[i,j] = deltas[2]

        dpeak_low[i,j] = deltas[idlow[0]]
        dcentroid_low[i,j] = deltas[idlow[1]]
        dlinewidth_low[i,j] = deltas[idlow[2]]

        if is_dbell:
            lineslope_up[i,j] = coeff[3]
            lineslope_low[i,j] = coeff[idlow[3]]
            dlineslope_up[i,j] = deltas[3]
            dlineslope_low[i,j] = deltas[idlow[3]]
        
    print ('Fitting two-component function along velocity axis of the input cube...')

    #********
    #MAKE FIT
    for i in range(nx):
        for j in range(ny):
            tmp_data = data[:,i,j]

            if mask[i,j]:
                n_mask += 1
                n_fit[i,j] = -10                                    
                continue
            
            try: 
                coeff, var_matrix = curve_fit(fit_func,
                                              vchannels, tmp_data,
                                              p0=pfunc_two(i,j),
                                              ftol=1e-10, xtol=1e-10, gtol=1e-10, method='lm')
                                              #bounds = [bound0, bound1])
                deltas = np.sqrt(np.abs(np.diag(var_matrix)))
                n_two += 1
                n_fit[i,j] = 2

                r"""
                if (deltas[:3]==np.inf).any():# or deltas[1] > 0.5*np.abs(dv):
                    coeff, var_matrix = curve_fit(fit_func1d,
                                                  vchannels, tmp_data,
                                                  p0=pfunc_one(i,j)
                    )
                    coeff = np.append(coeff, coeff)
                    deltas = np.sqrt(np.abs(np.diag(var_matrix)))
                    deltas = np.append(deltas, deltas)
                    n_one += 1
                    n_fit[i,j] = 1
                else:                
                    n_two += 1
                    n_fit[i,j] = 2
                #"""
                
            except RuntimeError:
                try: 
                    coeff, var_matrix = curve_fit(fit_func1d,
                                                  vchannels, tmp_data,
                                                  p0=pfunc_one(i,j)
                    )
                    coeff = np.append(coeff, coeff)
                    deltas = np.sqrt(np.abs(np.diag(var_matrix)))
                    deltas = np.append(deltas, deltas)
                    n_one += 1
                    n_fit[i,j] = 1
                    
                except RuntimeError:
                    n_bad += 1
                    n_fit[i,j] = 0                    
                    continue

            r"""
            if coeff[0] < coeff[idlow[0]]:
                coeff = np.append(coeff[idlow[0]:idlow[-1]+1], coeff[0:idlow[0]])
                deltas = np.append(deltas[idlow[0]:idlow[-1]+1], deltas[0:idlow[0]])
            #"""
            
            r"""
            #Can induce biases when upper and lower velocities are too close
            if np.abs(coeff[idlow[1]]-vel_peak_upper[i,j]) < np.abs(coeff[1]-vel_peak_upper[i,j]):
                coeff = np.append(coeff[idlow[0]:idlow[-1]+1], coeff[0:idlow[0]])
                deltas = np.append(deltas[idlow[0]:idlow[-1]+1], deltas[0:idlow[0]])
            #"""
            r"""
            #Sometimes the fit can output upper profiles near to zero intensity and in turn the lower profile is attributed to the bulk of the emission.
            #The following seems to improve things due to this in a few pixels. However, the conditional above accounts for this for the most part already.
            if coeff[0]<=1.0 and coeff[3]>1.0:
                coeff = np.append(coeff[idlow[0]:idlow[-1]+1], coeff[0:idlow[0]])
                deltas = np.append(deltas[idlow[0]:idlow[-1]+1], deltas[0:idlow[0]])
            #""" 

            fill_props(i,j,coeff)
                        
        _progress_bar(int(100*i/nx))
    _progress_bar(100)

    #*****************************
    #KEEP TRACK OF 'HOT' PIXELS
    #*****************************
    #Hot pixels will be tagged as -1

    def clean_nfit():
        mm = n_fit == -10 #noise
        ii = ((peak_up < 0.0) | (peak_low < 0.0)) & (~mm) #negative intensities
        peak_thres = 2*np.nanmax(cube_max)
        jj = ((peak_up > peak_thres) | (peak_low > peak_thres)) & (~mm) #too large intensities
        cc = (centroid_up == centroid_low) & (n_fit != 1) & (~mm) #up==low
        ww = ((np.abs(linewidth_up) <= 0.5*np.abs(dv)) | (np.abs(linewidth_low) <= 0.5*np.abs(dv))) & (~mm) #narrow component
        dd = ((np.abs(linewidth_up) > 5.0) | (np.abs(linewidth_low) > 5.0)) & (~mm) #Unrealistically broad component

        bb = False #dcentroid_up > 1*np.abs(dv)

        w3 = False #((np.abs(linewidth_up) <= 1*np.abs(dv)) | (np.abs(linewidth_low) <= 1*np.abs(dv))) & (~mm) #narrow component
        w4 = False #(np.abs(linewidth_low) >= 2*np.abs(linewidth_up)) & (~mm) #lower line width much larger than upper
        
        n_fit[ii+jj+cc+ww+dd] = -1 #bad pixel
        n_hot = np.sum(ii+jj+cc+ww+dd + w3+w4)

        n_fit[w3] = 3 #1chan narrow component
        n_fit[w4] = 4 #1chan narrow component
        
        return n_hot
        
    #**********
    #PACK PROPS
    #**********
    upper = [peak_up, centroid_up, linewidth_up]
    dupper = [dpeak_up, dcentroid_up, dlinewidth_up]
    lower = [peak_low, centroid_low, linewidth_low]
    dlower = [dpeak_low, dcentroid_low, dlinewidth_low]

    if is_dbell:
        upper += [lineslope_up]
        dupper += [dlineslope_up]
        lower += [lineslope_low]
        dlower += [dlineslope_low]

    n_hot = clean_nfit()
    
    print ('\nTwo-component fit did not converge for %.2f%s of the pixels'%(100.0*(n_bad)/(nx*ny),'%'))
    print ('A single component was fit for %.2f%s of the pixels'%(100.0*(n_one)/(nx*ny),'%'))
    print ('Masked pixels below intensity threshold: %.2f%s'%(100.0*(n_mask)/(nx*ny),'%'))
    print ('Hot pixels: %.2f%s'%(100.0*(n_hot)/(nx*ny),'%'))        

    if niter>0:
        print ('\nRe-doing fit for  %d hot pixels and %d single-component pixels'%(n_hot, n_one))

        def neighbour_guess(i, j, n_fit, neighs=3, av_func=np.nanmean):

            if i<neighs or j<neighs:
                return None
            
            neigh_arr = np.arange(-neighs, neighs+1)

            ileft, iright = i-neighs, i+neighs+1
            jleft, jright = j-neighs, j+neighs+1
            window = n_fit[ileft:iright, jleft:jright]

            masked = window==-10
            n_mask = np.sum(masked) #n masked pixels

            hot = window==-1
            n_hot = np.sum(hot) #n hot pixels
            
            one = window==1
            n_one = np.sum(one) #single-component pixels

            
            three = window==3
            n_three = np.sum(three) #narrow components

            four = window==4
            n_four = np.sum(four) #broad lower linewidth
            
            
            n_bad = n_mask+n_hot+n_one + n_three+n_four
            tot = (2*neighs+1)**2

            if tot-n_bad < n_bad:
                return None

            ic, jc = (~masked & ~hot & ~one  &  ~three & ~four).nonzero() #get clean pixels where double fit worked

            up_guess = [av_func(up_prop[ileft:iright, jleft:jright][ic, jc]) for up_prop in upper]
            low_guess = [av_func(low_prop[ileft:iright, jleft:jright][ic, jc]) for low_prop in lower]
            return up_guess+low_guess
                         
        for ni in range(niter):
            print ('Iteration #%d...'%(ni+1))

            #Select hot pixels (-1 flag) and single-component pixels (+1 flag)
            m, n = np.where((n_fit==-1) | (n_fit==1) | (n_fit>2))
            #m, n = np.where(n_fit==-1)
                
            for k in range(len(m)):
                i, j = m[k], n[k]                                
                uplow = neighbour_guess(i, j, n_fit, neighs=neighs, av_func=av_func)
                
                if uplow is None:
                    continue

                tmp_data = data[:,i,j]

                try:                    
                    coeff, var_matrix = curve_fit(fit_func,
                                                  vchannels, tmp_data,
                                                  p0 = uplow,
                                                  ftol=1e-10, xtol=1e-10, gtol=1e-10, method='lm')
                    deltas = np.sqrt(np.abs(np.diag(var_matrix)))
                    n_two += 1
                    n_fit[i,j] = 2
                    
                    r"""
                    if (deltas[:3]==np.inf).any(): # or deltas[1] > 0.5*np.abs(dv):
                        coeff, var_matrix = curve_fit(fit_func1d,
                                                      vchannels, tmp_data,
                                                      p0=pfunc_one(i,j)
                        )
                        coeff = np.append(coeff, coeff)
                        deltas = np.sqrt(np.abs(np.diag(var_matrix)))
                        deltas = np.append(deltas, deltas)
                        n_one += 1
                        n_fit[i,j] = 1
                    else:                
                        n_two += 1
                        n_fit[i,j] = 2
                    #"""
                except RuntimeError:
                    try: 
                        coeff, var_matrix = curve_fit(fit_func1d,
                                                      vchannels, tmp_data,
                                                      p0=pfunc_one(i,j)
                        )
                        coeff = np.append(coeff, coeff)
                        deltas = np.sqrt(np.abs(np.diag(var_matrix)))
                        deltas = np.append(deltas, deltas)
                        n_one += 1
                        n_fit[i,j] = 1
                    
                    except RuntimeError:
                        n_bad += 1
                        n_fit[i,j] = 0                    
                        continue

                fill_props(i,j,coeff)
                
            n_hot = clean_nfit()
            print ('Resulting hot pixels:', n_hot)
    
    return upper, dupper, lower, dlower, n_fit


def fit_gaussian(*args, **kwargs): #Backcompat
    return fit_onecomponent(*args, **kwargs)

def fit_onecomponent(
        cube, method='gaussian', lw_chans=1.0, peak_kernel=True, sigma_fit=None, sigma_thres=4, fit_continuum=False
):
    """
    Fit 'gaussian' or 'bell' profiles along the velocity axis of the input cube.
    
    Parameters
    ----------
    lw_chans : int, optional
        Number of channel widths to take as initial guess for line width

    sigma_fit : array_like, optional 
        2-D array of weights computed for each pixel. Shape must be equal to that of the spatial axes of the input data. 

    peak_kernel : bool, optional
        If True (default) the returned amplitude is the peak of the kernel fitted to the line. Otherwise, the actual peak of the line profile is returned.

    Returns
    -------
    A,B,C,dA,dB,dC : 2-D arrays
        Amplitude, centre, standard deviation, and their corresponding errors, obtained from Gaussians fitted to the input datacube along its velocity axis, on each spatial pixel.
    
    """    
    data = cube.data
    vchannels = cube.vchannels

    nchan, nx, ny = np.shape(data)
    n_fit = np.zeros((nx, ny))                        
    n_bad, n_mask = 0, 0
    
    peak, dpeak = np.zeros((nx, ny)), np.zeros((nx, ny))
    centroid, dcent = np.zeros((nx, ny)), np.zeros((nx, ny))
    linewidth, dlinew = np.zeros((nx, ny)), np.zeros((nx, ny))
    lineslope, dlineslope = np.zeros((nx, ny)), np.zeros((nx, ny))    
    if fit_continuum:
        continuum, dcontinuum = np.zeros((nx, ny)), np.zeros((nx, ny))
        
    is_bell = method=='bell'
    
    ind_max = np.nanargmax(data, axis=0)
    I_max = np.nanmax(data, axis=0)
    vel_peak = vchannels[ind_max]
    dv = lw_chans*np.mean(vchannels[1:]-vchannels[:-1])
    Ls = 2.0 #p0 Line slope
    
    if sigma_fit is None:
        sigma_func = lambda i,j: None
    else:
        sigma_func = lambda i,j: sigma_fit[:,i,j]

    #******************************
    #KERNELS AND RELEVANT FUNCTIONS
    if method in ['gaussian', 'gauss']:
        fit_func1d = _gauss
        if fit_continuum:
            pfunc_one = lambda i,j: [I_max[i,j], vel_peak[i,j], dv, 0.0]
        else:
            pfunc_one = lambda i,j: [I_max[i,j], vel_peak[i,j], dv]

    elif method=='bell':
        fit_func1d = _bell
        if fit_continuum:
            pfunc_one = lambda i,j: [I_max[i,j], vel_peak[i,j], dv, Ls, 0.0]
        else:
            pfunc_one = lambda i,j: [I_max[i,j], vel_peak[i,j], dv, Ls]
    else:
        _not_available(method)

    noise = np.std( np.append(data[:5,:,:], data[-5:,:,:], axis=0), axis=0) #rms intensity from first and last 5 channels
    mask = np.nanmax(data, axis=0) <= sigma_thres*noise
        
    print ('Fitting one-component function along velocity axis of the input cube...')

    for i in range(nx):
        for j in range(ny):
            tmp_data = data[:,i,j]
            
            if mask[i,j]:
                n_mask += 1
                n_fit[i,j] = -10                                    
                continue
            
            try:
                coeff, var_matrix = curve_fit(fit_func1d, vchannels, tmp_data, p0=pfunc_one(i,j), sigma=sigma_func(i,j))
                n_fit[i,j] = 1
                    
            except RuntimeError: 
                n_bad+=1
                n_fit[i,j] = 0                
                continue

            deltas = np.sqrt(np.diag(var_matrix))
            
            if peak_kernel:
                peak[i,j] = coeff[0]
            else:
                peak[i,j] = I_max[i,j]
                
            centroid[i,j] = coeff[1]
            linewidth[i,j] = coeff[2]
            dpeak[i,j], dcent[i,j], dlinew[i,j] = deltas[:3]

            if is_bell:
                lineslope[i,j] = coeff[3]
                dlineslope[i,j] = deltas[3]

            if fit_continuum:
                continuum[i,j] = coeff[-1]
                dcontinuum[i,j] = deltas[-1]
                
        _progress_bar(int(100*i/nx))

    _progress_bar(100)

    #***************
    #PACK AND RETURN
    #***************    
    upper = [peak, centroid, np.abs(linewidth)]
    dupper = [dpeak, dcent, dlinew]

    if is_bell:
        upper += [lineslope]
        dupper += [dlineslope]

    if fit_continuum:
        upper += [continuum]
        dupper += [dcontinuum]
        
    print ('\nOne-component fit did not converge for %.2f%s of the pixels'%(100.0*n_bad/(nx*ny),'%'))
    print ('Masked pixels below intensity threshold: %.2f%s'%(100.0*(n_mask)/(nx*ny),'%'))
    
    return upper, dupper, n_fit



from discminer.plottools import use_discminer_style, make_up_ax
from utils import init_data_and_model

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter

from astropy import units as u

import json
import copy
import sys

use_discminer_style()

with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']

Rmax = 1.1*best['intensity']['Rout']*u.au #Max model radius, 10% larger than disc Rout

#*******************
#LOAD DATA AND MODEL
#*******************
datacube, model = init_data_and_model()
model.make_model()

#**************
#MAKE PLOT
#**************
fig = plt.figure(figsize=(14,5))
ax0 = fig.add_axes([0.08,0.1,0.4,0.4])
ax1 = fig.add_axes([0.08,0.5,0.4,0.4])
ax2 = fig.add_axes([0.52,0.5,0.4,0.4])
ax3 = fig.add_axes([0.52,0.1,0.4,0.4])

#UPPER SURFACE
R_profile = np.linspace(0*u.au, Rmax, 100)
coords = {'R': R_profile}

linewidth_upper = model.get_attribute_map(coords, 'linewidth', surface='upper') #Computes coords{'z'} if not found.
intensity_upper = model.get_attribute_map(coords, 'intensity', surface='upper')
velocity_upper = model.get_attribute_map(coords, 'velocity', surface='upper') * model.params['velocity']['vel_sign']
z_upper = coords['z']*u.m.to('au')

kwargs_plot = dict(lw=3.5, c='tomato')
ax0.plot(R_profile, z_upper, label='Upper surface', **kwargs_plot)
ax1.plot(R_profile, velocity_upper, **kwargs_plot)
ax2.plot(R_profile, intensity_upper, **kwargs_plot)
ax3.plot(R_profile, linewidth_upper, **kwargs_plot)

make_up_ax(ax0, ylims=(0.0, 1.1*np.nanmax(z_upper)), labelbottom=True, labeltop=False, labelsize=11)
make_up_ax(ax1, labelbottom=False, labeltop=False, labelsize=11)
make_up_ax(ax2, labelbottom=False, labeltop=False, labelleft=False, labelright=True, labelsize=11)
make_up_ax(ax3, ylims=(0.1, 1.3), labelbottom=True, labeltop=False, labelleft=False, labelright=True, labelsize=11)

ax0.set_xlabel('Radius [au]')
ax0.set_ylabel('Elevation [au]', fontsize=12)

ax1.set_ylabel('Rot. Velocity [km/s]', fontsize=12)
ax2.set_ylabel('Peak Intensity []', fontsize=12, labelpad=25, rotation=-90)
ax2.yaxis.set_label_position('right')

ax3.set_xlabel('Radius [au]')
ax3.set_ylabel('Linewidth [km/s]', fontsize=12, labelpad=25, rotation=-90)
ax3.yaxis.set_label_position('right')

for axi in [ax0, ax1, ax2, ax3]:
    axi.yaxis.set_major_formatter(FormatStrFormatter('%4.1f'))

#LOWER SURFACE
coords = {'R': R_profile}    
linewidth_lower = model.get_attribute_map(coords, 'linewidth', surface='lower')
intensity_lower = model.get_attribute_map(coords, 'intensity', surface='lower')
velocity_lower = model.get_attribute_map(coords, 'velocity', surface='lower') * model.params['velocity']['vel_sign']
z_lower = -coords['z']*u.m.to('au')

kwargs_plot = dict(lw=3.5, ls=':', c='dodgerblue')
ax0.plot(R_profile, z_lower, label='Lower surface', **kwargs_plot)
ax1.plot(R_profile, velocity_lower, **kwargs_plot)
ax2.plot(R_profile, intensity_lower, **kwargs_plot)
ax3.plot(R_profile, linewidth_lower, **kwargs_plot)

ax0.legend(frameon=False)
fig.suptitle('Model Attributes', fontsize=18)

plt.savefig('model_attributes.png', bbox_inches='tight', dpi=200)
plt.show()
